import flax.linen as nn
from typing import Sequence, Callable, Union, Dict, Any, ClassVar
import jax.numpy as jnp
import jax
import numpy as np
from typing import Optional, Tuple
import numba
import dataclasses
from functools import partial

from flax.core.frozen_dict import FrozenDict


from ..utils.activations import chain,safe_sqrt
from ..utils import deep_update, mask_filter_1d
from ..utils.kspace import get_reciprocal_space_parameters
from .misc.misc import SwitchFunction
from ..utils.periodic_table import PERIODIC_TABLE, PERIODIC_TABLE_REV_IDX,CHEMICAL_BLOCKS,CHEMICAL_BLOCKS_NAMES


@dataclasses.dataclass(frozen=True)
class GraphGenerator:
    """Generate a graph from a set of coordinates

    FPID: GRAPH

    For now, we generate all pairs of atoms and filter based on cutoff.
    If a `nblist_skin` is present in the state, we generate a second graph with a larger cutoff that includes all pairs within the cutoff+skin. This graph is then reused by the `update_skin` method to update the original graph without recomputing the full nblist.
    """

    cutoff: float
    """Cutoff distance for the graph."""
    graph_key: str = "graph"
    """Key of the graph in the outputs."""
    switch_params: dict = dataclasses.field(default_factory=dict, hash=False)
    """Parameters for the switching function. See `fennol.models.misc.misc.SwitchFunction`."""
    kmax: int = 30
    """Maximum number of k-points to consider."""
    kthr: float = 1e-6
    """Threshold for k-point filtering."""
    k_space: bool = False
    """Whether to generate k-space information for the graph."""
    mult_size: float = 1.05
    """Multiplicative factor for resizing the nblist."""
    # covalent_cutoff: bool = False

    FPID: ClassVar[str] = "GRAPH"

    def init(self):
        return FrozenDict(
            {
                "max_nat": 1,
                "npairs": 1,
                "nblist_mult_size": self.mult_size,
            }
        )

    def get_processor(self) -> Tuple[nn.Module, Dict]:
        return GraphProcessor, {
            "cutoff": self.cutoff,
            "graph_key": self.graph_key,
            "switch_params": self.switch_params,
            "name": f"{self.graph_key}_Processor",
        }

    def get_graph_properties(self):
        return {
            self.graph_key: {
                "cutoff": self.cutoff,
                "directed": True,
            }
        }

    def __call__(self, state, inputs, return_state_update=False, add_margin=False):
        """build a nblist on cpu with numpy and dynamic shapes + store max shapes"""
        if self.graph_key in inputs:
            graph = inputs[self.graph_key]
            if "keep_graph" in graph:
                return state, inputs

        coords = np.array(inputs["coordinates"], dtype=np.float32)
        natoms = np.array(inputs["natoms"], dtype=np.int32)
        batch_index = np.array(inputs["batch_index"], dtype=np.int32)

        new_state = {**state}
        state_up = {}

        mult_size = state.get("nblist_mult_size", self.mult_size)
        assert mult_size >= 1.0, "mult_size should be larger or equal than 1.0"

        if natoms.shape[0] == 1:
            max_nat = coords.shape[0]
            true_max_nat = max_nat
        else:
            max_nat = state.get("max_nat", round(coords.shape[0] / natoms.shape[0]))
            true_max_nat = int(np.max(natoms))
            if true_max_nat > max_nat:
                add_atoms = state.get("add_atoms", 0)
                new_maxnat = true_max_nat + add_atoms
                state_up["max_nat"] = (new_maxnat, max_nat)
                new_state["max_nat"] = new_maxnat

        cutoff_skin = self.cutoff + state.get("nblist_skin", 0.0)

        ### compute indices of all pairs
        p1, p2 = np.triu_indices(true_max_nat, 1)
        p1, p2 = p1.astype(np.int32), p2.astype(np.int32)
        pbc_shifts = None
        if natoms.shape[0] > 1:
            ## batching => mask irrelevant pairs
            mask_p12 = (
                (p1[None, :] < natoms[:, None]) * (p2[None, :] < natoms[:, None])
            ).flatten()
            shift = np.concatenate(
                (np.array([0], dtype=np.int32), np.cumsum(natoms[:-1], dtype=np.int32))
            )
            p1 = np.where(mask_p12, (p1[None, :] + shift[:, None]).flatten(), -1)
            p2 = np.where(mask_p12, (p2[None, :] + shift[:, None]).flatten(), -1)

        apply_pbc = "cells" in inputs
        if not apply_pbc:
            ### NO PBC
            vec = coords[p2] - coords[p1]
        else:
            cells = np.array(inputs["cells"], dtype=np.float32)
            reciprocal_cells = np.array(inputs["reciprocal_cells"], dtype=np.float32)
            minimage = state.get("minimum_image", True)
            if minimage:
                ## MINIMUM IMAGE CONVENTION
                vec = coords[p2] - coords[p1]
                if cells.shape[0] == 1:
                    vecpbc = np.dot(vec, reciprocal_cells[0])
                    pbc_shifts = -np.round(vecpbc)
                    vec = vec + np.dot(pbc_shifts, cells[0])
                else:
                    batch_index_vec = batch_index[p1]
                    vecpbc = np.einsum(
                        "aj,aji->ai", vec, reciprocal_cells[batch_index_vec]
                    )
                    pbc_shifts = -np.round(vecpbc)
                    vec = vec + np.einsum(
                        "aj,aji->ai", pbc_shifts, cells[batch_index_vec]
                    )
            else:
                ### GENERAL PBC
                ## put all atoms in central box
                if cells.shape[0] == 1:
                    coords_pbc = np.dot(coords, reciprocal_cells[0])
                    at_shifts = -np.floor(coords_pbc)
                    coords_pbc = coords + np.dot(at_shifts, cells[0])
                else:
                    coords_pbc = np.einsum(
                        "aj,aji->ai", coords, reciprocal_cells[batch_index]
                    )
                    at_shifts = -np.floor(coords_pbc)
                    coords_pbc = coords + np.einsum(
                        "aj,aji->ai", at_shifts, cells[batch_index]
                    )
                vec = coords_pbc[p2] - coords_pbc[p1]

                ## compute maximum number of repeats
                inv_distances = (np.sum(reciprocal_cells**2, axis=1)) ** 0.5
                cdinv = cutoff_skin * inv_distances
                num_repeats_all = np.ceil(cdinv).astype(np.int32)
                if "true_sys" in inputs:
                    num_repeats_all = np.where(np.array(inputs["true_sys"],dtype=bool)[:, None], num_repeats_all, 0)
                # num_repeats_all = np.where(cdinv < 0.5, 0, num_repeats_all)
                num_repeats = np.max(num_repeats_all, axis=0)
                num_repeats_prev = np.array(state.get("num_repeats_pbc", (0, 0, 0)))
                if np.any(num_repeats > num_repeats_prev):
                    num_repeats_new = np.maximum(num_repeats, num_repeats_prev)
                    state_up["num_repeats_pbc"] = (
                        tuple(num_repeats_new),
                        tuple(num_repeats_prev),
                    )
                    new_state["num_repeats_pbc"] = tuple(num_repeats_new)
                ## build all possible shifts
                cell_shift_pbc = np.array(
                    np.meshgrid(*[np.arange(-n, n + 1) for n in num_repeats]),
                    dtype=cells.dtype,
                ).T.reshape(-1, 3)
                ## shift applied to vectors
                if cells.shape[0] == 1:
                    dvec = np.dot(cell_shift_pbc, cells[0])[None, :, :]
                    vec = (vec[:, None, :] + dvec).reshape(-1, 3)
                    pbc_shifts = np.broadcast_to(
                        cell_shift_pbc[None, :, :],
                        (p1.shape[0], cell_shift_pbc.shape[0], 3),
                    ).reshape(-1, 3)
                    p1 = np.broadcast_to(
                        p1[:, None], (p1.shape[0], cell_shift_pbc.shape[0])
                    ).flatten()
                    p2 = np.broadcast_to(
                        p2[:, None], (p2.shape[0], cell_shift_pbc.shape[0])
                    ).flatten()
                    if natoms.shape[0] > 1:
                        mask_p12 = np.broadcast_to(
                            mask_p12[:, None],
                            (mask_p12.shape[0], cell_shift_pbc.shape[0]),
                        ).flatten()
                else:
                    dvec = np.einsum("bj,sji->sbi", cell_shift_pbc, cells)

                    ## get pbc shifts specific to each box
                    cell_shift_pbc = np.broadcast_to(
                        cell_shift_pbc[None, :, :],
                        (num_repeats_all.shape[0], cell_shift_pbc.shape[0], 3),
                    )
                    mask = np.all(
                        np.abs(cell_shift_pbc) <= num_repeats_all[:, None, :], axis=-1
                    ).flatten()
                    idx = np.nonzero(mask)[0]
                    nshifts = idx.shape[0]
                    nshifts_prev = state.get("nshifts_pbc", 0)
                    if nshifts > nshifts_prev or add_margin:
                        nshifts_new = int(mult_size * max(nshifts, nshifts_prev)) + 1
                        state_up["nshifts_pbc"] = (nshifts_new, nshifts_prev)
                        new_state["nshifts_pbc"] = nshifts_new

                    dvec_filter = dvec.reshape(-1, 3)[idx, :]
                    cell_shift_pbc_filter = cell_shift_pbc.reshape(-1, 3)[idx, :]

                    ## get batch shift in the dvec_filter array
                    nrep = np.prod(2 * num_repeats_all + 1, axis=1)
                    bshift = np.concatenate((np.array([0]), np.cumsum(nrep)[:-1]))

                    ## compute vectors
                    batch_index_vec = batch_index[p1]
                    nrep_vec = np.where(mask_p12,nrep[batch_index_vec],0)
                    vec = vec.repeat(nrep_vec, axis=0)
                    nvec_pbc = nrep_vec.sum() #vec.shape[0]
                    nvec_pbc_prev = state.get("nvec_pbc", 0)
                    if nvec_pbc > nvec_pbc_prev or add_margin:
                        nvec_pbc_new = int(mult_size * max(nvec_pbc, nvec_pbc_prev)) + 1
                        state_up["nvec_pbc"] = (nvec_pbc_new, nvec_pbc_prev)
                        new_state["nvec_pbc"] = nvec_pbc_new

                    # print("cpu: ", nvec_pbc, nvec_pbc_prev, nshifts, nshifts_prev)
                    ## get shift index
                    dshift = np.concatenate(
                        (np.array([0]), np.cumsum(nrep_vec)[:-1])
                    ).repeat(nrep_vec)
                    # ishift = np.arange(dshift.shape[0])-dshift
                    # bshift_vec_rep = bshift[batch_index_vec].repeat(nrep_vec)
                    icellshift = (
                        np.arange(dshift.shape[0])
                        - dshift
                        + bshift[batch_index_vec].repeat(nrep_vec)
                    )
                    # shift vectors
                    vec = vec + dvec_filter[icellshift]
                    pbc_shifts = cell_shift_pbc_filter[icellshift]

                    p1 = np.repeat(p1, nrep_vec)
                    p2 = np.repeat(p2, nrep_vec)
                    if natoms.shape[0] > 1:
                        mask_p12 = np.repeat(mask_p12, nrep_vec)

        ## compute distances
        d12 = (vec**2).sum(axis=-1)
        if natoms.shape[0] > 1:
            d12 = np.where(mask_p12, d12, cutoff_skin**2)

        ## filter pairs
        max_pairs = state.get("npairs", 1)
        mask = d12 < cutoff_skin**2
        idx = np.nonzero(mask)[0]
        npairs = idx.shape[0]
        if npairs > max_pairs or add_margin:
            prev_max_pairs = max_pairs
            max_pairs = int(mult_size * max(npairs, max_pairs)) + 1
            state_up["npairs"] = (max_pairs, prev_max_pairs)
            new_state["npairs"] = max_pairs

        nat = coords.shape[0]
        edge_src = np.full(max_pairs, nat, dtype=np.int32)
        edge_dst = np.full(max_pairs, nat, dtype=np.int32)
        d12_ = np.full(max_pairs, cutoff_skin**2)
        edge_src[:npairs] = p1[idx]
        edge_dst[:npairs] = p2[idx]
        d12_[:npairs] = d12[idx]
        d12 = d12_

        if apply_pbc:
            pbc_shifts_ = np.zeros((max_pairs, 3))
            pbc_shifts_[:npairs] = pbc_shifts[idx]
            pbc_shifts = pbc_shifts_
            if not minimage:
                pbc_shifts[:npairs] = (
                    pbc_shifts[:npairs]
                    + at_shifts[edge_dst[:npairs]]
                    - at_shifts[edge_src[:npairs]]
                )

        if "nblist_skin" in state:
            edge_src_skin = edge_src
            edge_dst_skin = edge_dst
            if apply_pbc:
                pbc_shifts_skin = pbc_shifts
            max_pairs_skin = state.get("npairs_skin", 1)
            mask = d12 < self.cutoff**2
            idx = np.nonzero(mask)[0]
            npairs_skin = idx.shape[0]
            if npairs_skin > max_pairs_skin or add_margin:
                prev_max_pairs_skin = max_pairs_skin
                max_pairs_skin = int(mult_size * max(npairs_skin, max_pairs_skin)) + 1
                state_up["npairs_skin"] = (max_pairs_skin, prev_max_pairs_skin)
                new_state["npairs_skin"] = max_pairs_skin
            edge_src = np.full(max_pairs_skin, nat, dtype=np.int32)
            edge_dst = np.full(max_pairs_skin, nat, dtype=np.int32)
            d12_ = np.full(max_pairs_skin, self.cutoff**2)
            edge_src[:npairs_skin] = edge_src_skin[idx]
            edge_dst[:npairs_skin] = edge_dst_skin[idx]
            d12_[:npairs_skin] = d12[idx]
            d12 = d12_
            if apply_pbc:
                pbc_shifts = np.full((max_pairs_skin, 3), 0.0)
                pbc_shifts[:npairs_skin] = pbc_shifts_skin[idx]

        ## symmetrize
        edge_src, edge_dst = np.concatenate((edge_src, edge_dst)), np.concatenate(
            (edge_dst, edge_src)
        )
        d12 = np.concatenate((d12, d12))
        if apply_pbc:
            pbc_shifts = np.concatenate((pbc_shifts, -pbc_shifts))

        graph = inputs.get(self.graph_key, {})
        graph_out = {
            **graph,
            "edge_src": edge_src,
            "edge_dst": edge_dst,
            "d12": d12,
            "overflow": False,
            "pbc_shifts": pbc_shifts,
        }
        if "nblist_skin" in state:
            graph_out["edge_src_skin"] = edge_src_skin
            graph_out["edge_dst_skin"] = edge_dst_skin
            if apply_pbc:
                graph_out["pbc_shifts_skin"] = pbc_shifts_skin

        if self.k_space and apply_pbc:
            if "k_points" not in graph:
                ks, _, _, bewald = get_reciprocal_space_parameters(
                    reciprocal_cells, self.cutoff, self.kmax, self.kthr
                )
            graph_out["k_points"] = ks
            graph_out["b_ewald"] = bewald

        output = {**inputs, self.graph_key: graph_out}

        if return_state_update:
            return FrozenDict(new_state), output, state_up
        return FrozenDict(new_state), output

    def check_reallocate(self, state, inputs, parent_overflow=False):
        """check for overflow and reallocate nblist if necessary"""
        overflow = parent_overflow or inputs[self.graph_key].get("overflow", False)
        if not overflow:
            return state, {}, inputs, False

        add_margin = inputs[self.graph_key].get("overflow", False)
        state, inputs, state_up = self(
            state, inputs, return_state_update=True, add_margin=add_margin
        )
        return state, state_up, inputs, True

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        """build a nblist on accelerator with jax and precomputed shapes"""
        if self.graph_key in inputs:
            graph = inputs[self.graph_key]
            if "keep_graph" in graph:
                return inputs
        coords = inputs["coordinates"]
        natoms = inputs["natoms"]
        batch_index = inputs["batch_index"]

        if natoms.shape[0] == 1:
            max_nat = coords.shape[0]
        else:
            max_nat = state.get(
                "max_nat", int(round(coords.shape[0] / natoms.shape[0]))
            )
        cutoff_skin = self.cutoff + state.get("nblist_skin", 0.0)

        ### compute indices of all pairs
        p1, p2 = np.triu_indices(max_nat, 1)
        p1, p2 = p1.astype(np.int32), p2.astype(np.int32)
        pbc_shifts = None
        if natoms.shape[0] > 1:
            ## batching => mask irrelevant pairs
            mask_p12 = (
                (p1[None, :] < natoms[:, None]) * (p2[None, :] < natoms[:, None])
            ).flatten()
            shift = jnp.concatenate(
                (jnp.array([0], dtype=jnp.int32), jnp.cumsum(natoms[:-1]))
            )
            p1 = jnp.where(mask_p12, (p1[None, :] + shift[:, None]).flatten(), -1)
            p2 = jnp.where(mask_p12, (p2[None, :] + shift[:, None]).flatten(), -1)

        ## compute vectors
        overflow_repeats = jnp.asarray(False, dtype=bool)
        if "cells" not in inputs:
            vec = coords[p2] - coords[p1]
        else:
            cells = inputs["cells"]
            reciprocal_cells = inputs["reciprocal_cells"]
            minimage = state.get("minimum_image", True)

            def compute_pbc(vec, reciprocal_cell, cell, mode="round"):
                vecpbc = jnp.dot(vec, reciprocal_cell)
                if mode == "round":
                    pbc_shifts = -jnp.round(vecpbc)
                elif mode == "floor":
                    pbc_shifts = -jnp.floor(vecpbc)
                else:
                    raise NotImplementedError(f"Unknown mode {mode} for compute_pbc.")
                return vec + jnp.dot(pbc_shifts, cell), pbc_shifts

            if minimage:
                ## minimum image convention
                vec = coords[p2] - coords[p1]

                if cells.shape[0] == 1:
                    vec, pbc_shifts = compute_pbc(vec, reciprocal_cells[0], cells[0])
                else:
                    batch_index_vec = batch_index[p1]
                    vec, pbc_shifts = jax.vmap(compute_pbc)(
                        vec, reciprocal_cells[batch_index_vec], cells[batch_index_vec]
                    )
            else:
                ### general PBC only for single cell yet
                # if cells.shape[0] > 1:
                #     raise NotImplementedError(
                #         "General PBC not implemented for batches on accelerator."
                #     )
                # cell = cells[0]
                # reciprocal_cell = reciprocal_cells[0]

                ## put all atoms in central box
                if cells.shape[0] == 1:
                    coords_pbc, at_shifts = compute_pbc(
                        coords, reciprocal_cells[0], cells[0], mode="floor"
                    )
                else:
                    coords_pbc, at_shifts = jax.vmap(
                        partial(compute_pbc, mode="floor")
                    )(coords, reciprocal_cells[batch_index], cells[batch_index])
                vec = coords_pbc[p2] - coords_pbc[p1]
                num_repeats = state.get("num_repeats_pbc", (0, 0, 0))
                # if num_repeats is None:
                #     raise ValueError(
                #         "num_repeats_pbc should be provided for general PBC on accelerator. Call the numpy routine (self.__call__) first."
                #     )
                # check if num_repeats is larger than previous
                inv_distances = jnp.linalg.norm(reciprocal_cells, axis=1)
                cdinv = cutoff_skin * inv_distances
                num_repeats_all = jnp.ceil(cdinv).astype(jnp.int32)
                if "true_sys" in inputs:
                    num_repeats_all = jnp.where(inputs["true_sys"][:,None], num_repeats_all, 0)
                num_repeats_new = jnp.max(num_repeats_all, axis=0)
                overflow_repeats = jnp.any(num_repeats_new > jnp.asarray(num_repeats))

                cell_shift_pbc = jnp.asarray(
                    np.array(
                        np.meshgrid(*[np.arange(-n, n + 1) for n in num_repeats]),
                        dtype=cells.dtype,
                    ).T.reshape(-1, 3)
                )

                if cells.shape[0] == 1:
                    vec = (vec[:,None,:] + jnp.dot(cell_shift_pbc, cells[0])[None, :, :]).reshape(-1, 3)    
                    pbc_shifts = jnp.broadcast_to(
                        cell_shift_pbc[None, :, :],
                        (p1.shape[0], cell_shift_pbc.shape[0], 3),
                    ).reshape(-1, 3)
                    p1 = jnp.broadcast_to(
                        p1[:, None], (p1.shape[0], cell_shift_pbc.shape[0])
                    ).flatten()
                    p2 = jnp.broadcast_to(
                        p2[:, None], (p2.shape[0], cell_shift_pbc.shape[0])
                    ).flatten()
                    if natoms.shape[0] > 1:
                        mask_p12 = jnp.broadcast_to(
                            mask_p12[:, None], (mask_p12.shape[0], cell_shift_pbc.shape[0])
                        ).flatten()
                else:
                    dvec = jnp.einsum("bj,sji->sbi", cell_shift_pbc, cells).reshape(-1, 3)

                    ## get pbc shifts specific to each box
                    cell_shift_pbc = jnp.broadcast_to(
                        cell_shift_pbc[None, :, :],
                        (num_repeats_all.shape[0], cell_shift_pbc.shape[0], 3),
                    )
                    mask = jnp.all(
                        jnp.abs(cell_shift_pbc) <= num_repeats_all[:, None, :], axis=-1
                    ).flatten()
                    max_shifts  = state.get("nshifts_pbc", 1)

                    cell_shift_pbc = cell_shift_pbc.reshape(-1,3)
                    shiftx,shifty,shiftz = cell_shift_pbc[:,0],cell_shift_pbc[:,1],cell_shift_pbc[:,2]
                    dvecx,dvecy,dvecz = dvec[:,0],dvec[:,1],dvec[:,2]
                    (dvecx, dvecy,dvecz,shiftx,shifty,shiftz), scatter_idx, nshifts = mask_filter_1d(
                        mask,
                        max_shifts,
                        (dvecx, 0.),
                        (dvecy, 0.),
                        (dvecz, 0.),
                        (shiftx, 0),
                        (shifty, 0),
                        (shiftz, 0),
                    )
                    dvec = jnp.stack((dvecx,dvecy,dvecz),axis=-1)
                    cell_shift_pbc = jnp.stack((shiftx,shifty,shiftz),axis=-1)
                    overflow_repeats = overflow_repeats | (nshifts > max_shifts)

                    ## get batch shift in the dvec_filter array
                    nrep = jnp.prod(2 * num_repeats_all + 1, axis=1)
                    bshift = jnp.concatenate((jnp.array([0],dtype=jnp.int32), jnp.cumsum(nrep)[:-1]))

                    ## repeat vectors
                    nvec_max = state.get("nvec_pbc", 1)
                    batch_index_vec = batch_index[p1]
                    nrep_vec = jnp.where(mask_p12,nrep[batch_index_vec],0)
                    nvec = nrep_vec.sum()
                    overflow_repeats = overflow_repeats | (nvec > nvec_max)
                    vec = jnp.repeat(vec,nrep_vec,axis=0,total_repeat_length=nvec_max)
                    # jax.debug.print("{nvec} {nvec_max} {nshifts} {max_shifts}",nvec=nvec,nvec_max=jnp.asarray(nvec_max),nshifts=nshifts,max_shifts=jnp.asarray(max_shifts))

                    ## get shift index
                    dshift = jnp.concatenate(
                        (jnp.array([0],dtype=jnp.int32), jnp.cumsum(nrep_vec)[:-1])
                    )
                    if nrep_vec.size == 0:
                        dshift = jnp.array([],dtype=jnp.int32)
                    dshift = jnp.repeat(dshift,nrep_vec, total_repeat_length=nvec_max)
                    bshift = jnp.repeat(bshift[batch_index_vec],nrep_vec, total_repeat_length=nvec_max)
                    icellshift = jnp.arange(dshift.shape[0]) - dshift + bshift
                    vec = vec + dvec[icellshift]
                    pbc_shifts = cell_shift_pbc[icellshift]
                    p1 = jnp.repeat(p1,nrep_vec, total_repeat_length=nvec_max)
                    p2 = jnp.repeat(p2,nrep_vec, total_repeat_length=nvec_max)
                    if natoms.shape[0] > 1:
                        mask_p12 = jnp.repeat(mask_p12,nrep_vec, total_repeat_length=nvec_max)
                

        ## compute distances
        d12 = (vec**2).sum(axis=-1)
        if natoms.shape[0] > 1:
            d12 = jnp.where(mask_p12, d12, cutoff_skin**2)

        ## filter pairs
        max_pairs = state.get("npairs", 1)
        mask = d12 < cutoff_skin**2
        (edge_src, edge_dst, d12), scatter_idx, npairs = mask_filter_1d(
            mask,
            max_pairs,
            (jnp.asarray(p1, dtype=jnp.int32), coords.shape[0]),
            (jnp.asarray(p2, dtype=jnp.int32), coords.shape[0]),
            (d12, cutoff_skin**2),
        )
        if "cells" in inputs:
            pbc_shifts = (
                jnp.full((max_pairs, 3), 0.0, dtype=pbc_shifts.dtype)
                .at[scatter_idx]
                .set(pbc_shifts, mode="drop")
            )
            if not minimage:
                pbc_shifts = (
                    pbc_shifts
                    + at_shifts.at[edge_dst].get(fill_value=0.0)
                    - at_shifts.at[edge_src].get(fill_value=0.0)
                )

        ## check for overflow
        if natoms.shape[0] == 1:
            true_max_nat = coords.shape[0]
        else:
            true_max_nat = jnp.max(natoms)
        overflow_count = npairs > max_pairs
        overflow_at = true_max_nat > max_nat
        overflow = overflow_count | overflow_at | overflow_repeats

        if "nblist_skin" in state:
            # edge_mask_skin = edge_mask
            edge_src_skin = edge_src
            edge_dst_skin = edge_dst
            if "cells" in inputs:
                pbc_shifts_skin = pbc_shifts
            max_pairs_skin = state.get("npairs_skin", 1)
            mask = d12 < self.cutoff**2
            (edge_src, edge_dst, d12), scatter_idx, npairs_skin = mask_filter_1d(
                mask,
                max_pairs_skin,
                (edge_src, coords.shape[0]),
                (edge_dst, coords.shape[0]),
                (d12, self.cutoff**2),
            )
            if "cells" in inputs:
                pbc_shifts = (
                    jnp.full((max_pairs_skin, 3), 0.0, dtype=pbc_shifts.dtype)
                    .at[scatter_idx]
                    .set(pbc_shifts, mode="drop")
                )
            overflow = overflow | (npairs_skin > max_pairs_skin)

        ## symmetrize
        edge_src, edge_dst = jnp.concatenate((edge_src, edge_dst)), jnp.concatenate(
            (edge_dst, edge_src)
        )
        d12 = jnp.concatenate((d12, d12))
        if "cells" in inputs:
            pbc_shifts = jnp.concatenate((pbc_shifts, -pbc_shifts))

        graph = inputs[self.graph_key] if self.graph_key in inputs else {}
        graph_out = {
            **graph,
            "edge_src": edge_src,
            "edge_dst": edge_dst,
            "d12": d12,
            "overflow": overflow,
            "pbc_shifts": pbc_shifts,
        }
        if "nblist_skin" in state:
            graph_out["edge_src_skin"] = edge_src_skin
            graph_out["edge_dst_skin"] = edge_dst_skin
            if "cells" in inputs:
                graph_out["pbc_shifts_skin"] = pbc_shifts_skin

        if self.k_space and "cells" in inputs:
            if "k_points" not in graph:
                raise NotImplementedError(
                    "k_space generation not implemented on accelerator. Call the numpy routine (self.__call__) first."
                )
        return {**inputs, self.graph_key: graph_out}

    @partial(jax.jit, static_argnums=(0,))
    def update_skin(self, inputs):
        """update the nblist without recomputing the full nblist"""
        graph = inputs[self.graph_key]

        edge_src_skin = graph["edge_src_skin"]
        edge_dst_skin = graph["edge_dst_skin"]
        coords = inputs["coordinates"]
        vec = coords.at[edge_dst_skin].get(
            mode="fill", fill_value=self.cutoff
        ) - coords.at[edge_src_skin].get(mode="fill", fill_value=0.0)

        if "cells" in inputs:
            pbc_shifts_skin = graph["pbc_shifts_skin"]
            cells = inputs["cells"]
            if cells.shape[0] == 1:
                vec = vec + jnp.dot(pbc_shifts_skin, cells[0])
            else:
                batch_index_vec = inputs["batch_index"][edge_src_skin]
                vec = vec + jax.vmap(jnp.dot)(pbc_shifts_skin, cells[batch_index_vec])

        nat = coords.shape[0]
        d12 = jnp.sum(vec**2, axis=-1)
        mask = d12 < self.cutoff**2
        max_pairs = graph["edge_src"].shape[0] // 2
        (edge_src, edge_dst, d12), scatter_idx, npairs = mask_filter_1d(
            mask,
            max_pairs,
            (edge_src_skin, nat),
            (edge_dst_skin, nat),
            (d12, self.cutoff**2),
        )
        if "cells" in inputs:
            pbc_shifts = (
                jnp.full((max_pairs, 3), 0.0, dtype=pbc_shifts_skin.dtype)
                .at[scatter_idx]
                .set(pbc_shifts_skin)
            )

        overflow = graph.get("overflow", False) | (npairs > max_pairs)
        graph_out = {
            **graph,
            "edge_src": jnp.concatenate((edge_src, edge_dst)),
            "edge_dst": jnp.concatenate((edge_dst, edge_src)),
            "d12": jnp.concatenate((d12, d12)),
            "overflow": overflow,
        }
        if "cells" in inputs:
            graph_out["pbc_shifts"] = jnp.concatenate((pbc_shifts, -pbc_shifts))

        if self.k_space and "cells" in inputs:
            if "k_points" not in graph:
                raise NotImplementedError(
                    "k_space generation not implemented on accelerator. Call the numpy routine (self.__call__) first."
                )

        return {**inputs, self.graph_key: graph_out}


class GraphProcessor(nn.Module):
    """Process a pre-generated graph

    The pre-generated graph should contain the following keys:
    - edge_src: source indices of the edges
    - edge_dst: destination indices of the edges
    - pbcs_shifts: pbc shifts for the edges (only if `cells` are present in the inputs)

    This module is automatically added to a FENNIX model when a GraphGenerator is used.

    """

    cutoff: float
    """Cutoff distance for the graph."""
    graph_key: str = "graph"
    """Key of the graph in the outputs."""
    switch_params: dict = dataclasses.field(default_factory=dict)
    """Parameters for the switching function. See `fennol.models.misc.misc.SwitchFunction`."""

    @nn.compact
    def __call__(self, inputs: Union[dict, Tuple[jax.Array, dict]]):
        graph = inputs[self.graph_key]
        coords = inputs["coordinates"]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        # edge_mask = edge_src < coords.shape[0]
        vec = coords.at[edge_dst].get(mode="fill", fill_value=self.cutoff) - coords.at[
            edge_src
        ].get(mode="fill", fill_value=0.0)
        if "cells" in inputs:
            cells = inputs["cells"]
            if cells.shape[0] == 1:
                vec = vec + jnp.dot(graph["pbc_shifts"], cells[0])
            else:
                batch_index_vec = inputs["batch_index"][edge_src]
                vec = vec + jax.vmap(jnp.dot)(
                    graph["pbc_shifts"], cells[batch_index_vec]
                )

        d2  = jnp.sum(vec**2, axis=-1)
        distances = safe_sqrt(d2)
        edge_mask = distances < self.cutoff

        switch = SwitchFunction(
            **{**self.switch_params, "cutoff": self.cutoff, "graph_key": None}
        )((distances, edge_mask))

        graph_out = {
            **graph,
            "vec": vec,
            "distances": distances,
            "switch": switch,
            "edge_mask": edge_mask,
        }

        if "alch_group" in inputs:
            alch_group = inputs["alch_group"]
            lambda_e = inputs["alch_elambda"]
            mask = alch_group[edge_src] == alch_group[edge_dst]
            graph_out["switch_raw"] = switch
            graph_out["switch"] = jnp.where(
                mask,
                switch,
                0.5*(1.-jnp.cos(jnp.pi*lambda_e)) * switch ,
            )

            if "alch_softcore_e" in inputs or "alch_softcore_v" in inputs:
                graph_out["distances_raw"] = distances
                if "alch_softcore_e" in inputs:
                    alch_alpha = (1-inputs["alch_elambda"])*inputs["alch_softcore_e"]**2
                else:
                    alch_alpha = (1-inputs["alch_vlambda"])*inputs["alch_softcore_v"]**2
                distances = jnp.where(
                    mask,
                    distances,
                    safe_sqrt(alch_alpha + d2 * (1. - alch_alpha/self.cutoff**2))
                )  
                graph_out["distances"] = distances

        return {**inputs, self.graph_key: graph_out}


@dataclasses.dataclass(frozen=True)
class GraphFilter:
    """Filter a graph based on a cutoff distance

    FPID: GRAPH_FILTER
    """

    cutoff: float
    """Cutoff distance for the filtering."""
    parent_graph: str
    """Key of the parent graph in the inputs."""
    graph_key: str
    """Key of the filtered graph in the outputs."""
    remove_hydrogens: int = False
    """Remove edges where the source is a hydrogen atom."""
    switch_params: FrozenDict = dataclasses.field(default_factory=FrozenDict)
    """Parameters for the switching function. See `fennol.models.misc.misc.SwitchFunction`."""
    k_space: bool = False
    """Generate k-space information for the graph."""
    kmax: int = 30
    """Maximum number of k-points to consider."""
    kthr: float = 1e-6
    """Threshold for k-point filtering."""
    mult_size: float = 1.05
    """Multiplicative factor for resizing the nblist."""

    FPID: ClassVar[str] = "GRAPH_FILTER"

    def init(self):
        return FrozenDict(
            {
                "npairs": 1,
                "nblist_mult_size": self.mult_size,
            }
        )

    def get_processor(self) -> Tuple[nn.Module, Dict]:
        return GraphFilterProcessor, {
            "cutoff": self.cutoff,
            "graph_key": self.graph_key,
            "parent_graph": self.parent_graph,
            "name": f"{self.graph_key}_Filter_{self.parent_graph}",
            "switch_params": self.switch_params,
        }

    def get_graph_properties(self):
        return {
            self.graph_key: {
                "cutoff": self.cutoff,
                "directed": True,
                "parent_graph": self.parent_graph,
            }
        }

    def __call__(self, state, inputs, return_state_update=False, add_margin=False):
        """filter a nblist on cpu with numpy and dynamic shapes + store max shapes"""
        graph_in = inputs[self.parent_graph]
        nat = inputs["species"].shape[0]

        new_state = {**state}
        state_up = {}
        mult_size = state.get("nblist_mult_size", self.mult_size)
        assert mult_size >= 1., "nblist_mult_size should be >= 1."

        edge_src = np.array(graph_in["edge_src"], dtype=np.int32)
        d12 = np.array(graph_in["d12"], dtype=np.float32)
        if self.remove_hydrogens:
            species = inputs["species"]
            src_idx = (edge_src < nat).nonzero()[0]
            mask = np.zeros(edge_src.shape[0], dtype=bool)
            mask[src_idx] = (species > 1)[edge_src[src_idx]]
            d12 = np.where(mask, d12, self.cutoff**2)
        mask = d12 < self.cutoff**2

        max_pairs = state.get("npairs", 1)
        idx = np.nonzero(mask)[0]
        npairs = idx.shape[0]
        if npairs > max_pairs or add_margin:
            prev_max_pairs = max_pairs
            max_pairs = int(mult_size * max(npairs, max_pairs)) + 1
            state_up["npairs"] = (max_pairs, prev_max_pairs)
            new_state["npairs"] = max_pairs

        filter_indices = np.full(max_pairs, edge_src.shape[0], dtype=np.int32)
        edge_src = np.full(max_pairs, nat, dtype=np.int32)
        edge_dst = np.full(max_pairs, nat, dtype=np.int32)
        d12_ = np.full(max_pairs, self.cutoff**2)
        filter_indices[:npairs] = idx
        edge_src[:npairs] = graph_in["edge_src"][idx]
        edge_dst[:npairs] = graph_in["edge_dst"][idx]
        d12_[:npairs] = d12[idx]
        d12 = d12_

        graph = inputs[self.graph_key] if self.graph_key in inputs else {}
        graph_out = {
            **graph,
            "edge_src": edge_src,
            "edge_dst": edge_dst,
            "filter_indices": filter_indices,
            "d12": d12,
            "overflow": False,
        }

        if self.k_space and "cells" in inputs:
            if "k_points" not in graph:
                ks, _, _, bewald = get_reciprocal_space_parameters(
                    inputs["reciprocal_cells"], self.cutoff, self.kmax, self.kthr
                )
            graph_out["k_points"] = ks
            graph_out["b_ewald"] = bewald

        output = {**inputs, self.graph_key: graph_out}
        if return_state_update:
            return FrozenDict(new_state), output, state_up
        return FrozenDict(new_state), output

    def check_reallocate(self, state, inputs, parent_overflow=False):
        """check for overflow and reallocate nblist if necessary"""
        overflow = parent_overflow or inputs[self.graph_key].get("overflow", False)
        if not overflow:
            return state, {}, inputs, False

        add_margin = inputs[self.graph_key].get("overflow", False)
        state, inputs, state_up = self(
            state, inputs, return_state_update=True, add_margin=add_margin
        )
        return state, state_up, inputs, True

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        """filter a nblist on accelerator with jax and precomputed shapes"""
        graph_in = inputs[self.parent_graph]
        if state is None:
            # skin update mode
            graph = inputs[self.graph_key]
            max_pairs = graph["edge_src"].shape[0]
        else:
            max_pairs = state.get("npairs", 1)

        max_pairs_in = graph_in["edge_src"].shape[0]
        nat = inputs["species"].shape[0]

        edge_src = graph_in["edge_src"]
        d12 = graph_in["d12"]
        if self.remove_hydrogens:
            species = inputs["species"]
            mask = (species > 1)[edge_src]
            d12 = jnp.where(mask, d12, self.cutoff**2)
        mask = d12 < self.cutoff**2

        (edge_src, edge_dst, d12, filter_indices), _, npairs = mask_filter_1d(
            mask,
            max_pairs,
            (edge_src, nat),
            (graph_in["edge_dst"], nat),
            (d12, self.cutoff**2),
            (jnp.arange(max_pairs_in, dtype=jnp.int32), max_pairs_in),
        )

        graph = inputs[self.graph_key] if self.graph_key in inputs else {}
        overflow = graph.get("overflow", False) | (npairs > max_pairs)
        graph_out = {
            **graph,
            "edge_src": edge_src,
            "edge_dst": edge_dst,
            "filter_indices": filter_indices,
            "d12": d12,
            "overflow": overflow,
        }

        if self.k_space and "cells" in inputs:
            if "k_points" not in graph:
                raise NotImplementedError(
                    "k_space generation not implemented on accelerator. Call the numpy routine (self.__call__) first."
                )

        return {**inputs, self.graph_key: graph_out}

    @partial(jax.jit, static_argnums=(0,))
    def update_skin(self, inputs):
        return self.process(None, inputs)


class GraphFilterProcessor(nn.Module):
    """Filter processing for a pre-generated graph

    This module is automatically added to a FENNIX model when a GraphFilter is used.
    """

    cutoff: float
    """Cutoff distance for the filtering."""
    graph_key: str
    """Key of the filtered graph in the inputs."""
    parent_graph: str
    """Key of the parent graph in the inputs."""
    switch_params: dict = dataclasses.field(default_factory=dict)
    """Parameters for the switching function. See `fennol.models.misc.misc.SwitchFunction`."""

    @nn.compact
    def __call__(self, inputs: Union[dict, Tuple[jax.Array, dict]]):
        graph_in = inputs[self.parent_graph]
        graph = inputs[self.graph_key]

        d_key = "distances_raw" if "distances_raw" in graph else "distances"

        if graph_in["vec"].shape[0] == 0:
            vec = graph_in["vec"]
            distances = graph_in[d_key]
            filter_indices = jnp.asarray([], dtype=jnp.int32)
        else:
            filter_indices = graph["filter_indices"]
            vec = (
                graph_in["vec"]
                .at[filter_indices]
                .get(mode="fill", fill_value=self.cutoff)
            )
            distances = (
                graph_in[d_key]
                .at[filter_indices]
                .get(mode="fill", fill_value=self.cutoff)
            )

        edge_mask = distances < self.cutoff
        switch = SwitchFunction(
            **{**self.switch_params, "cutoff": self.cutoff, "graph_key": None}
        )((distances, edge_mask))

        graph_out = {
            **graph,
            "vec": vec,
            "distances": distances,
            "switch": switch,
            "filter_indices": filter_indices,
            "edge_mask": edge_mask,
        }

        if "alch_group" in inputs:
            edge_src=graph["edge_src"]
            edge_dst=graph["edge_dst"]
            alch_group = inputs["alch_group"]
            lambda_e = inputs["alch_elambda"]
            mask = alch_group[edge_src] == alch_group[edge_dst]
            graph_out["switch_raw"] = switch
            graph_out["switch"] = jnp.where(
                mask,
                switch,
                0.5*(1.-jnp.cos(jnp.pi*lambda_e)) * switch ,
            )

            if "alch_softcore_e" in inputs or "alch_softcore_v" in inputs:
                graph_out["distances_raw"] = distances
                if "alch_softcore_e" in inputs:
                    alch_alpha = (1-inputs["alch_elambda"])*inputs["alch_softcore_e"]**2
                else:
                    alch_alpha = (1-inputs["alch_vlambda"])*inputs["alch_softcore_v"]**2
                distances = jnp.where(
                    mask,
                    distances,
                    safe_sqrt(alch_alpha + distances**2 * (1. - alch_alpha/self.cutoff**2))
                )  
                graph_out["distances"] = distances

        return {**inputs, self.graph_key: graph_out}


@dataclasses.dataclass(frozen=True)
class GraphAngularExtension:
    """Add angles list to a graph

    FPID: GRAPH_ANGULAR_EXTENSION
    """

    mult_size: float = 1.05
    """Multiplicative factor for resizing the nblist."""
    add_neigh: int = 5
    """Additional neighbors to add to the nblist when resizing."""
    graph_key: str = "graph"
    """Key of the graph in the inputs."""

    FPID: ClassVar[str] = "GRAPH_ANGULAR_EXTENSION"

    def init(self):
        return FrozenDict(
            {
                "nangles": 0,
                "nblist_mult_size": self.mult_size,
                "max_neigh": self.add_neigh,
                "add_neigh": self.add_neigh,
            }
        )

    def get_processor(self) -> Tuple[nn.Module, Dict]:
        return GraphAngleProcessor, {
            "graph_key": self.graph_key,
            "name": f"{self.graph_key}_AngleProcessor",
        }

    def get_graph_properties(self):
        return {
            self.graph_key: {
                "has_angles": True,
            }
        }

    def __call__(self, state, inputs, return_state_update=False, add_margin=False):
        """build angle nblist on cpu with numpy and dynamic shapes + store max shapes"""
        graph = inputs[self.graph_key]
        edge_src = np.array(graph["edge_src"], dtype=np.int32)

        new_state = {**state}
        state_up = {}
        mult_size = state.get("nblist_mult_size", self.mult_size)
        assert mult_size >= 1., "nblist_mult_size should be >= 1."

        ### count number of neighbors
        nat = inputs["species"].shape[0]
        count = np.zeros(nat + 1, dtype=np.int32)
        np.add.at(count, edge_src, 1)
        max_count = int(np.max(count[:-1]))

        ### get sizes
        max_neigh = state.get("max_neigh", self.add_neigh)
        nedge = edge_src.shape[0]
        if max_count > max_neigh or add_margin:
            prev_max_neigh = max_neigh
            max_neigh = max(max_count, max_neigh) + state.get(
                "add_neigh", self.add_neigh
            )
            state_up["max_neigh"] = (max_neigh, prev_max_neigh)
            new_state["max_neigh"] = max_neigh

        max_neigh_arr = np.empty(max_neigh, dtype=bool)

        nedge = edge_src.shape[0]

        ### sort edge_src
        idx_sort = np.argsort(edge_src)
        edge_src_sorted = edge_src[idx_sort]

        ### map sparse to dense nblist
        offset = np.tile(np.arange(max_count), nat)
        if max_count * nat >= nedge:
            offset = np.tile(np.arange(max_count), nat)[:nedge]
        else:
            offset = np.zeros(nedge, dtype=np.int32)
            offset[: max_count * nat] = np.tile(np.arange(max_count), nat)

        # offset = jnp.where(edge_src_sorted < nat, offset, 0)
        mask = edge_src_sorted < nat
        indices = edge_src_sorted * max_count + offset
        indices = indices[mask]
        idx_sort = idx_sort[mask]
        edge_idx = np.full(nat * max_count, nedge, dtype=np.int32)
        edge_idx[indices] = idx_sort
        edge_idx = edge_idx.reshape(nat, max_count)

        ### find all triplet for each atom center
        local_src, local_dst = np.triu_indices(max_count, 1)
        angle_src = edge_idx[:, local_src].flatten()
        angle_dst = edge_idx[:, local_dst].flatten()

        ### mask for valid angles
        mask1 = angle_src < nedge
        mask2 = angle_dst < nedge
        angle_mask = mask1 & mask2

        max_angles = state.get("nangles", 0)
        idx = np.nonzero(angle_mask)[0]
        nangles = idx.shape[0]
        if nangles > max_angles or add_margin:
            max_angles_prev = max_angles
            max_angles = int(mult_size * max(nangles, max_angles)) + 1
            state_up["nangles"] = (max_angles, max_angles_prev)
            new_state["nangles"] = max_angles

        ## filter angles to sparse representation
        angle_src_ = np.full(max_angles, nedge, dtype=np.int32)
        angle_dst_ = np.full(max_angles, nedge, dtype=np.int32)
        angle_src_[:nangles] = angle_src[idx]
        angle_dst_[:nangles] = angle_dst[idx]

        central_atom = np.full(max_angles, nat, dtype=np.int32)
        central_atom[:nangles] = edge_src[angle_src_[:nangles]]

        ## update graph
        output = {
            **inputs,
            self.graph_key: {
                **graph,
                "angle_src": angle_src_,
                "angle_dst": angle_dst_,
                "central_atom": central_atom,
                "angle_overflow": False,
                "max_neigh": max_neigh,
                "__max_neigh_array": max_neigh_arr,
            },
        }

        if return_state_update:
            return FrozenDict(new_state), output, state_up
        return FrozenDict(new_state), output

    def check_reallocate(self, state, inputs, parent_overflow=False):
        """check for overflow and reallocate nblist if necessary"""
        overflow = parent_overflow or inputs[self.graph_key]["angle_overflow"]
        if not overflow:
            return state, {}, inputs, False

        add_margin = inputs[self.graph_key]["angle_overflow"]
        state, inputs, state_up = self(
            state, inputs, return_state_update=True, add_margin=add_margin
        )
        return state, state_up, inputs, True

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        """build angle nblist on accelerator with jax and precomputed shapes"""
        graph = inputs[self.graph_key]
        edge_src = graph["edge_src"]

        ### count number of neighbors
        nat = inputs["species"].shape[0]
        count = jnp.zeros(nat, dtype=jnp.int32).at[edge_src].add(1, mode="drop")
        max_count = jnp.max(count)

        ### get sizes
        if state is None:
            max_neigh_arr = graph["__max_neigh_array"]
            max_neigh = max_neigh_arr.shape[0]
            prev_nangles = graph["angle_src"].shape[0]
        else:
            max_neigh = state.get("max_neigh", self.add_neigh)
            max_neigh_arr = jnp.empty(max_neigh, dtype=bool)
            prev_nangles = state.get("nangles", 0)

        nedge = edge_src.shape[0]

        ### sort edge_src
        idx_sort = jnp.argsort(edge_src).astype(jnp.int32)
        edge_src_sorted = edge_src[idx_sort]

        ### map sparse to dense nblist
        if max_neigh * nat < nedge:
            raise ValueError("Found max_neigh*nat < nedge. This should not happen.")
        offset = jnp.asarray(
            np.tile(np.arange(max_neigh), nat)[:nedge], dtype=jnp.int32
        )
        # offset = jnp.where(edge_src_sorted < nat, offset, 0)
        indices = edge_src_sorted * max_neigh + offset
        edge_idx = (
            jnp.full(nat * max_neigh, nedge, dtype=jnp.int32)
            .at[indices]
            .set(idx_sort, mode="drop")
            .reshape(nat, max_neigh)
        )

        ### find all triplet for each atom center
        local_src, local_dst = np.triu_indices(max_neigh, 1)
        angle_src = edge_idx[:, local_src].flatten()
        angle_dst = edge_idx[:, local_dst].flatten()

        ### mask for valid angles
        mask1 = angle_src < nedge
        mask2 = angle_dst < nedge
        angle_mask = mask1 & mask2

        ## filter angles to sparse representation
        (angle_src, angle_dst), _, nangles = mask_filter_1d(
            angle_mask,
            prev_nangles,
            (angle_src, nedge),
            (angle_dst, nedge),
        )
        ## find central atom
        central_atom = edge_src[angle_src]

        ## check for overflow
        angle_overflow = nangles > prev_nangles
        neigh_overflow = max_count > max_neigh
        overflow = graph.get("angle_overflow", False) | angle_overflow | neigh_overflow

        ## update graph
        output = {
            **inputs,
            self.graph_key: {
                **graph,
                "angle_src": angle_src,
                "angle_dst": angle_dst,
                "central_atom": central_atom,
                "angle_overflow": overflow,
                # "max_neigh": max_neigh,
                "__max_neigh_array": max_neigh_arr,
            },
        }

        return output

    @partial(jax.jit, static_argnums=(0,))
    def update_skin(self, inputs):
        return self.process(None, inputs)


class GraphAngleProcessor(nn.Module):
    """Process a pre-generated graph to compute angles

    This module is automatically added to a FENNIX model when a GraphAngularExtension is used.

    """

    graph_key: str
    """Key of the graph in the inputs."""

    @nn.compact
    def __call__(self, inputs: Union[dict, Tuple[jax.Array, dict]]):
        graph = inputs[self.graph_key]
        distances = graph["distances_raw"] if "distances_raw" in graph else graph["distances"]
        vec = graph["vec"]
        angle_src = graph["angle_src"]
        angle_dst = graph["angle_dst"]

        dir = vec / jnp.clip(distances[:, None], min=1.0e-5)
        cos_angles = (
            dir.at[angle_src].get(mode="fill", fill_value=0.5)
            * dir.at[angle_dst].get(mode="fill", fill_value=0.5)
        ).sum(axis=-1)

        angles = jnp.arccos(0.95 * cos_angles)

        return {
            **inputs,
            self.graph_key: {
                **graph,
                # "cos_angles": cos_angles,
                "angles": angles,
                # "angle_mask": angle_mask,
            },
        }


@dataclasses.dataclass(frozen=True)
class SpeciesIndexer:
    """Build an index that splits atomic arrays by species.

    FPID: SPECIES_INDEXER

    If `species_order` is specified, the output will be a dense array with size (len(species_order), max_size) that can directly index atomic arrays.
    If `species_order` is None, the output will be a dictionary with species as keys and an index to filter atomic arrays for that species as values.

    """

    output_key: str = "species_index"
    """Key for the output dictionary."""
    species_order: Optional[str] = None
    """Comma separated list of species in the order they should be indexed."""
    add_atoms: int = 0
    """Additional atoms to add to the sizes."""
    add_atoms_margin: int = 10
    """Additional atoms to add to the sizes when adding margin."""

    FPID: ClassVar[str] = "SPECIES_INDEXER"

    def init(self):
        return FrozenDict(
            {
                "sizes": {},
            }
        )

    def __call__(self, state, inputs, return_state_update=False, add_margin=False):
        species = np.array(inputs["species"], dtype=np.int32)
        nat = species.shape[0]
        set_species, counts = np.unique(species, return_counts=True)

        new_state = {**state}
        state_up = {}

        sizes = state.get("sizes", FrozenDict({}))
        new_sizes = {**sizes}
        up_sizes = False
        counts_dict = {}
        for s, c in zip(set_species, counts):
            if s <= 0:
                continue
            counts_dict[s] = c
            if c > sizes.get(s, 0):
                up_sizes = True
                add_atoms = state.get("add_atoms", self.add_atoms)
                if add_margin:
                    add_atoms += state.get("add_atoms_margin", self.add_atoms_margin)
                new_sizes[s] = c + add_atoms

        new_sizes = FrozenDict(new_sizes)
        if up_sizes:
            state_up["sizes"] = (new_sizes, sizes)
            new_state["sizes"] = new_sizes

        if self.species_order is not None:
            species_order = [el.strip() for el in self.species_order.split(",")]
            max_size_prev = state.get("max_size", 0)
            max_size = max(new_sizes.values())
            if max_size > max_size_prev:
                state_up["max_size"] = (max_size, max_size_prev)
                new_state["max_size"] = max_size
                max_size_prev = max_size

            species_index = np.full((len(species_order), max_size), nat, dtype=np.int32)
            for i, el in enumerate(species_order):
                s = PERIODIC_TABLE_REV_IDX[el]
                if s in counts_dict.keys():
                    species_index[i, : counts_dict[s]] = np.nonzero(species == s)[0]
        else:
            species_index = {
                PERIODIC_TABLE[s]: np.full(c, nat, dtype=np.int32)
                for s, c in new_sizes.items()
            }
            for s, c in zip(set_species, counts):
                if s <= 0:
                    continue
                species_index[PERIODIC_TABLE[s]][:c] = np.nonzero(species == s)[0]

        output = {
            **inputs,
            self.output_key: species_index,
            self.output_key + "_overflow": False,
        }

        if return_state_update:
            return FrozenDict(new_state), output, state_up
        return FrozenDict(new_state), output

    def check_reallocate(self, state, inputs, parent_overflow=False):
        """check for overflow and reallocate nblist if necessary"""
        overflow = parent_overflow or inputs[self.output_key + "_overflow"]
        if not overflow:
            return state, {}, inputs, False

        add_margin = inputs[self.output_key + "_overflow"]
        state, inputs, state_up = self(
            state, inputs, return_state_update=True, add_margin=add_margin
        )
        return state, state_up, inputs, True
        # return state, {}, inputs, parent_overflow

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        # assert (
        #     self.output_key in inputs
        # ), f"Species Index {self.output_key} must be provided on accelerator. Call the numpy routine (self.__call__) first."

        recompute_species_index = "recompute_species_index" in inputs.get("flags", {})
        if self.output_key in inputs and not recompute_species_index:
            return inputs

        if state is None:
            raise ValueError("Species Indexer state must be provided on accelerator.")

        species = inputs["species"]
        nat = species.shape[0]

        sizes = state["sizes"]

        if self.species_order is not None:
            species_order = [el.strip() for el in self.species_order.split(",")]
            max_size = state["max_size"]

            species_index = jnp.full(
                (len(species_order), max_size), nat, dtype=jnp.int32
            )
            for i, el in enumerate(species_order):
                s = PERIODIC_TABLE_REV_IDX[el]
                if s in sizes.keys():
                    c = sizes[s]
                    species_index = species_index.at[i, :].set(
                        jnp.nonzero(species == s, size=max_size, fill_value=nat)[0]
                    )
                # if s in counts_dict.keys():
                #     species_index[i, : counts_dict[s]] = np.nonzero(species == s)[0]
        else:
            # species_index = {
            # PERIODIC_TABLE[s]: jnp.nonzero(species == s, size=c, fill_value=nat)[0]
            # for s, c in sizes.items()
            # }
            species_index = {}
            overflow = False
            natcount = 0
            for s, c in sizes.items():
                mask = species == s
                new_size = jnp.sum(mask)
                natcount = natcount + new_size
                overflow = overflow | (new_size > c)  # check if sizes are correct
                species_index[PERIODIC_TABLE[s]] = jnp.nonzero(
                    species == s, size=c, fill_value=nat
                )[0]

            mask = species <= 0
            new_size = jnp.sum(mask)
            natcount = natcount + new_size
            overflow = overflow | (
                natcount < species.shape[0]
            )  # check if any species missing

        return {
            **inputs,
            self.output_key: species_index,
            self.output_key + "_overflow": overflow,
        }

    @partial(jax.jit, static_argnums=(0,))
    def update_skin(self, inputs):
        return self.process(None, inputs)

@dataclasses.dataclass(frozen=True)
class BlockIndexer:
    """Build an index that splits atomic arrays by chemical blocks.

    FPID: BLOCK_INDEXER

    If `species_order` is specified, the output will be a dense array with size (len(species_order), max_size) that can directly index atomic arrays.
    If `species_order` is None, the output will be a dictionary with species as keys and an index to filter atomic arrays for that species as values.

    """

    output_key: str = "block_index"
    """Key for the output dictionary."""
    add_atoms: int = 0
    """Additional atoms to add to the sizes."""
    add_atoms_margin: int = 10
    """Additional atoms to add to the sizes when adding margin."""
    split_CNOPSSe: bool = False

    FPID: ClassVar[str] = "BLOCK_INDEXER"

    def init(self):
        return FrozenDict(
            {
                "sizes": {},
            }
        )

    def build_chemical_blocks(self):
        _CHEMICAL_BLOCKS_NAMES = CHEMICAL_BLOCKS_NAMES.copy()
        if self.split_CNOPSSe:
            _CHEMICAL_BLOCKS_NAMES[1] = "C"
            _CHEMICAL_BLOCKS_NAMES.extend(["N","O","P","S","Se"])
        _CHEMICAL_BLOCKS = CHEMICAL_BLOCKS.copy()
        if self.split_CNOPSSe:
            _CHEMICAL_BLOCKS[6] = 1
            _CHEMICAL_BLOCKS[7] = len(CHEMICAL_BLOCKS_NAMES)
            _CHEMICAL_BLOCKS[8] = len(CHEMICAL_BLOCKS_NAMES)+1
            _CHEMICAL_BLOCKS[15] = len(CHEMICAL_BLOCKS_NAMES)+2
            _CHEMICAL_BLOCKS[16] = len(CHEMICAL_BLOCKS_NAMES)+3
            _CHEMICAL_BLOCKS[34] = len(CHEMICAL_BLOCKS_NAMES)+4
        return _CHEMICAL_BLOCKS_NAMES, _CHEMICAL_BLOCKS

    def __call__(self, state, inputs, return_state_update=False, add_margin=False):
        _CHEMICAL_BLOCKS_NAMES, _CHEMICAL_BLOCKS = self.build_chemical_blocks()

        species = np.array(inputs["species"], dtype=np.int32)
        blocks = _CHEMICAL_BLOCKS[species]
        nat = species.shape[0]
        set_blocks, counts = np.unique(blocks, return_counts=True)

        new_state = {**state}
        state_up = {}

        sizes = state.get("sizes", FrozenDict({}))
        new_sizes = {**sizes}
        up_sizes = False
        for s, c in zip(set_blocks, counts):
            if s < 0:
                continue
            key = (s, _CHEMICAL_BLOCKS_NAMES[s])
            if c > sizes.get(key, 0):
                up_sizes = True
                add_atoms = state.get("add_atoms", self.add_atoms)
                if add_margin:
                    add_atoms += state.get("add_atoms_margin", self.add_atoms_margin)
                new_sizes[key] = c + add_atoms

        new_sizes = FrozenDict(new_sizes)
        if up_sizes:
            state_up["sizes"] = (new_sizes, sizes)
            new_state["sizes"] = new_sizes

        block_index = {n:None for n in _CHEMICAL_BLOCKS_NAMES}
        for (_,n), c in new_sizes.items():
            block_index[n] = np.full(c, nat, dtype=np.int32)
        # block_index = {
            # n: np.full(c, nat, dtype=np.int32)
            # for (_,n), c in new_sizes.items()
        # }
        for s, c in zip(set_blocks, counts):
            if s < 0:
                continue
            block_index[_CHEMICAL_BLOCKS_NAMES[s]][:c] = np.nonzero(blocks == s)[0]

        output = {
            **inputs,
            self.output_key: block_index,
            self.output_key + "_overflow": False,
        }

        if return_state_update:
            return FrozenDict(new_state), output, state_up
        return FrozenDict(new_state), output

    def check_reallocate(self, state, inputs, parent_overflow=False):
        """check for overflow and reallocate nblist if necessary"""
        overflow = parent_overflow or inputs[self.output_key + "_overflow"]
        if not overflow:
            return state, {}, inputs, False

        add_margin = inputs[self.output_key + "_overflow"]
        state, inputs, state_up = self(
            state, inputs, return_state_update=True, add_margin=add_margin
        )
        return state, state_up, inputs, True
        # return state, {}, inputs, parent_overflow

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        _CHEMICAL_BLOCKS_NAMES, _CHEMICAL_BLOCKS = self.build_chemical_blocks()
        # assert (
        #     self.output_key in inputs
        # ), f"Species Index {self.output_key} must be provided on accelerator. Call the numpy routine (self.__call__) first."

        recompute_species_index = "recompute_species_index" in inputs.get("flags", {})
        if self.output_key in inputs and not recompute_species_index:
            return inputs

        if state is None:
            raise ValueError("Block Indexer state must be provided on accelerator.")

        species = inputs["species"]
        blocks = jnp.asarray(_CHEMICAL_BLOCKS)[species]
        nat = species.shape[0]

        sizes = state["sizes"]

        # species_index = {
        # PERIODIC_TABLE[s]: jnp.nonzero(species == s, size=c, fill_value=nat)[0]
        # for s, c in sizes.items()
        # }
        block_index = {n: None for n in _CHEMICAL_BLOCKS_NAMES}
        overflow = False
        natcount = 0
        for (s,name), c in sizes.items():
            mask = blocks == s
            new_size = jnp.sum(mask)
            natcount = natcount + new_size
            overflow = overflow | (new_size > c)  # check if sizes are correct
            block_index[name] = jnp.nonzero(
                mask, size=c, fill_value=nat
            )[0]

        mask = blocks < 0
        new_size = jnp.sum(mask)
        natcount = natcount + new_size
        overflow = overflow | (
            natcount < species.shape[0]
        )  # check if any species missing

        return {
            **inputs,
            self.output_key: block_index,
            self.output_key + "_overflow": overflow,
        }

    @partial(jax.jit, static_argnums=(0,))
    def update_skin(self, inputs):
        return self.process(None, inputs)


@dataclasses.dataclass(frozen=True)
class AtomPadding:
    """Pad atomic arrays to a fixed size."""

    mult_size: float = 1.2
    """Multiplicative factor for resizing the atomic arrays."""
    add_sys: int = 0

    def init(self):
        return {"prev_nat": 0, "prev_nsys": 0}

    def __call__(self, state, inputs: Dict) -> Union[dict, jax.Array]:
        species = inputs["species"]
        nat = species.shape[0]

        prev_nat = state.get("prev_nat", 0)
        prev_nat_ = prev_nat
        if nat > prev_nat_:
            prev_nat_ = int(self.mult_size * nat) + 1

        nsys = len(inputs["natoms"])
        prev_nsys = state.get("prev_nsys", 0)
        prev_nsys_ = prev_nsys
        if nsys > prev_nsys_:
            prev_nsys_ = nsys + self.add_sys

        add_atoms = prev_nat_ - nat
        add_sys = prev_nsys_ - nsys  + 1
        output = {**inputs}
        if add_atoms > 0:
            for k, v in inputs.items():
                if isinstance(v, np.ndarray) or isinstance(v, jax.Array):
                    if v.shape[0] == nat:
                        output[k] = np.append(
                            v,
                            np.zeros((add_atoms, *v.shape[1:]), dtype=v.dtype),
                            axis=0,
                        )
                    elif v.shape[0] == nsys:
                        if k == "cells":
                            output[k] = np.append(
                                v,
                                1000
                                * np.eye(3, dtype=v.dtype)[None, :, :].repeat(
                                    add_sys, axis=0
                                ),
                                axis=0,
                            )
                        else:
                            output[k] = np.append(
                                v,
                                np.zeros((add_sys, *v.shape[1:]), dtype=v.dtype),
                                axis=0,
                            )
            output["natoms"] = np.append(
                inputs["natoms"], np.zeros(add_sys, dtype=np.int32)
            )
            output["species"] = np.append(
                species, -1 * np.ones(add_atoms, dtype=species.dtype)
            )
            output["batch_index"] = np.append(
                inputs["batch_index"],
                np.array([output["natoms"].shape[0] - 1] * add_atoms, dtype=inputs["batch_index"].dtype),
            )
            if "system_index" in inputs:
                output["system_index"] = np.append(
                    inputs["system_index"],
                    np.array([output["natoms"].shape[0] - 1] * add_sys, dtype=inputs["system_index"].dtype),
                )

        output["true_atoms"] = output["species"] > 0
        output["true_sys"] = np.arange(len(output["natoms"])) < nsys

        state = {**state, "prev_nat": prev_nat_, "prev_nsys": prev_nsys_}

        return FrozenDict(state), output


def atom_unpadding(inputs: Dict[str, Any]) -> Dict[str, Any]:
    """Remove padding from atomic arrays."""
    if "true_atoms" not in inputs:
        return inputs

    species = inputs["species"]
    true_atoms = inputs["true_atoms"]
    true_sys = inputs["true_sys"]
    natall = species.shape[0]
    nat = np.argmax(species <= 0)
    if nat == 0:
        return inputs

    natoms = inputs["natoms"]
    nsysall = len(natoms)

    output = {**inputs}
    for k, v in inputs.items():
        if isinstance(v, jax.Array) or isinstance(v, np.ndarray):
            if v.ndim == 0:
                continue
            if v.shape[0] == natall:
                output[k] = v[true_atoms]
            elif v.shape[0] == nsysall:
                output[k] = v[true_sys]
    del output["true_sys"]
    del output["true_atoms"]
    return output


def check_input(inputs):
    """Check the input dictionary for required keys and types."""
    assert "species" in inputs, "species must be provided"
    assert "coordinates" in inputs, "coordinates must be provided"
    species = inputs["species"].astype(np.int32)
    ifake = np.argmax(species <= 0)
    if ifake > 0:
        assert np.all(species[:ifake] > 0), "species must be positive"
    nat = inputs["species"].shape[0]

    natoms = inputs.get("natoms", np.array([nat], dtype=np.int32)).astype(np.int32)
    batch_index = inputs.get(
        "batch_index", np.repeat(np.arange(len(natoms), dtype=np.int32), natoms)
    ).astype(np.int32)
    output = {**inputs, "natoms": natoms, "batch_index": batch_index}
    if "cells" in inputs:
        cells = inputs["cells"]
        if "reciprocal_cells" not in inputs:
            reciprocal_cells = np.linalg.inv(cells)
        else:
            reciprocal_cells = inputs["reciprocal_cells"]
        if cells.ndim == 2:
            cells = cells[None, :, :]
        if reciprocal_cells.ndim == 2:
            reciprocal_cells = reciprocal_cells[None, :, :]
        output["cells"] = cells
        output["reciprocal_cells"] = reciprocal_cells

    return output


def convert_to_jax(data):
    """Convert a numpy arrays to jax arrays in a pytree."""

    def convert(x):
        if isinstance(x, np.ndarray):
            # if x.dtype == np.float64:
            #     return jnp.asarray(x, dtype=jnp.float32)
            return jnp.asarray(x)
        return x

    return jax.tree_util.tree_map(convert, data)


class JaxConverter(nn.Module):
    """Convert numpy arrays to jax arrays in a pytree."""

    def __call__(self, data):
        return convert_to_jax(data)


@dataclasses.dataclass(frozen=True)
class PreprocessingChain:
    """Chain of preprocessing layers."""

    layers: Tuple[Callable[..., Dict[str, Any]]]
    """Preprocessing layers."""
    use_atom_padding: bool = False
    """Add an AtomPadding layer at the beginning of the chain."""
    atom_padder: AtomPadding = AtomPadding()
    """AtomPadding layer."""

    def __post_init__(self):
        if not isinstance(self.layers, Sequence):
            raise ValueError(
                f"'layers' must be a sequence, got '{type(self.layers).__name__}'."
            )
        if not self.layers:
            raise ValueError(f"Error: no Preprocessing layers were provided.")

    def __call__(self, state, inputs: Dict[str, Any]) -> Dict[str, Any]:
        do_check_input = state.get("check_input", True)
        if do_check_input:
            inputs = check_input(inputs)
        new_state = []
        layer_state = state["layers_state"]
        i = 0
        if self.use_atom_padding:
            s, inputs = self.atom_padder(layer_state[0], inputs)
            new_state.append(s)
            i += 1
        for layer in self.layers:
            s, inputs = layer(layer_state[i], inputs, return_state_update=False)
            new_state.append(s)
            i += 1
        return FrozenDict({**state, "layers_state": tuple(new_state)}), convert_to_jax(
            inputs
        )

    def check_reallocate(self, state, inputs):
        new_state = []
        state_up = []
        layer_state = state["layers_state"]
        i = 0
        if self.use_atom_padding:
            new_state.append(layer_state[0])
            i += 1
        parent_overflow = False
        for layer in self.layers:
            s, s_up, inputs, parent_overflow = layer.check_reallocate(
                layer_state[i], inputs, parent_overflow
            )
            new_state.append(s)
            state_up.append(s_up)
            i += 1

        if not parent_overflow:
            return state, {}, inputs, False
        return (
            FrozenDict({**state, "layers_state": tuple(new_state)}),
            state_up,
            inputs,
            True,
        )

    def atom_padding(self, state, inputs):
        if self.use_atom_padding:
            padder_state = state["layers_state"][0]
            return self.atom_padder(padder_state, inputs)
        return state, inputs

    @partial(jax.jit, static_argnums=(0, 1))
    def process(self, state, inputs):
        layer_state = state["layers_state"]
        i = 1 if self.use_atom_padding else 0
        for layer in self.layers:
            inputs = layer.process(layer_state[i], inputs)
            i += 1
        return inputs

    @partial(jax.jit, static_argnums=(0))
    def update_skin(self, inputs):
        for layer in self.layers:
            inputs = layer.update_skin(inputs)
        return inputs

    def init(self):
        state = []
        if self.use_atom_padding:
            state.append(self.atom_padder.init())
        for layer in self.layers:
            state.append(layer.init())
        return FrozenDict({"check_input": True, "layers_state": state})

    def init_with_output(self, inputs):
        state = self.init()
        return self(state, inputs)

    def get_processors(self):
        processors = []
        for layer in self.layers:
            if hasattr(layer, "get_processor"):
                processors.append(layer.get_processor())
        return processors

    def get_graphs_properties(self):
        properties = {}
        for layer in self.layers:
            if hasattr(layer, "get_graph_properties"):
                properties = deep_update(properties, layer.get_graph_properties())
        return properties


# PREPROCESSING = {
#     "GRAPH": GraphGenerator,
#     # "GRAPH_FIXED": GraphGeneratorFixed,
#     "GRAPH_FILTER": GraphFilter,
#     "GRAPH_ANGULAR_EXTENSION": GraphAngularExtension,
#     # "GRAPH_DENSE_EXTENSION": GraphDenseExtension,
#     "SPECIES_INDEXER": SpeciesIndexer,
# }



from typing import Sequence, Tuple, Dict, Optional, Any
import flax.linen as nn
from inspect import isclass, ismodule
from pkgutil import iter_modules
import importlib
import importlib.util
import os
import glob

### python modules where FENNIX Modules are defined ###
from . import misc,physics, embeddings,preprocessing


MODULES: Dict[str, nn.Module] = {}
PREPROCESSING: Dict = {}


def available_fennix_modules():
    return list(MODULES.keys())

def available_fennix_preprocessing():
    return list(PREPROCESSING.keys())


def register_fennix_module(module: nn.Module, FID: Optional[str] = None):
    if FID is not None:
        names = [FID.upper()]
    else:
        if not hasattr(module, "FID"):
            print(f"Warning: module {module.__name__} does not have a FID field and no explicit FID was provided. Module was NOT registered.")
        if not (isinstance(
            module.FID, str
        ) or isinstance(module.FID,tuple)):
            print(f"Warning: module {module.__name__} has an invalid FID field. Module was NOT registered.")
        if isinstance(module.FID, str):
            names = [module.FID.upper()]
        else:
            names = [fid.upper() for fid in module.FID]
    for name in names:
        if name in MODULES and MODULES[name] != module:
            raise ValueError(
                f"A different module identified as '{name}' is already registered !"
            )
        MODULES[name] = module

def register_fennix_preprocessing(module, FPID: Optional[str] = None):
    if FPID is not None:
        names = [FPID.upper()]
    else:
        if not hasattr(module, "FPID"):
            print(f"Warning: module {module.__name__} does not have a FPID field and no explicit FPID was provided. Module was NOT registered.")
        if not (isinstance(
            module.FPID, str
        ) or isinstance(module.FPID,tuple)):
            print(f"Warning: module {module.__name__} has an invalid FPID field. Module was NOT registered.")
        if isinstance(module.FPID, str):
            names = [module.FPID.upper()]
        else:
            names = [fid.upper() for fid in module.FPID]
    for name in names:
        if name in PREPROCESSING and PREPROCESSING[name] != module:
            raise ValueError(
                f"A different module identified as '{name}' is already registered !"
            )
        PREPROCESSING[name] = module


def register_fennix_modules(module, recurs=0, max_recurs=2):
    if ismodule(module) and hasattr(module,"__path__"):
        for _, name, _ in iter_modules(module.__path__):
            sub_module = __import__(f"{module.__name__}.{name}", fromlist=[""])
            register_fennix_modules(sub_module, recurs=recurs + 1, max_recurs=max_recurs)
    for m in module.__dict__.values():
        if isclass(m) and issubclass(m, nn.Module) and m != nn.Module:
            if hasattr(m, "FID"):
                register_fennix_module(m)
        elif isclass(m) and hasattr(m, "FPID"):
            register_fennix_preprocessing(m)


### REGISTER DEFAULT MODULES #####################
for mods in [misc, physics, embeddings,preprocessing]:
    register_fennix_modules(mods)
module_path = os.environ.get("FENNOL_MODULES_PATH","").split(":")
for path in module_path:
    if os.path.exists(path):
        for file in glob.glob(f"{path}/*.py"):
            spec = importlib.util.spec_from_file_location("custom_modules", file)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            register_fennix_modules(module)

##################################################

def get_modules_documentation():
    doc = {}
    for name, module in MODULES.items():
        doc[name] = module.__doc__
    return doc


class FENNIXModules(nn.Module):
    r"""Sequential module that applies a sequence of FENNIX modules.

    Attributes:
        layers (Sequence[Tuple[nn.Module, Dict]]): Sequence of tuples (module, parameters) to apply.

    """

    layers: Sequence[Tuple[nn.Module, Dict]]

    def __post_init__(self):
        if not isinstance(self.layers, Sequence):
            raise ValueError(
                f"'layers' must be a sequence, got '{type(self.layers).__name__}'."
            )
        super().__post_init__()

    @nn.compact
    def __call__(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        if not self.layers:
            raise ValueError(f"Empty Sequential module {self.name}.")

        outputs = inputs
        for layer, prms in self.layers:
            outputs = layer(**prms)(outputs)
        return outputs



import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Union, ClassVar, Optional
import numpy as np
from ...utils.periodic_table import D3_COV_RADII
import dataclasses
from ..misc.encodings import RadialBasis, SpeciesEncoding
from ...utils.spherical_harmonics import generate_spherical_harmonics
from ..misc.e3 import  ChannelMixing
from ..misc.nets import FullyConnectedNet,BlockIndexNet
from ...utils.activations import activation_from_str
from ...utils import AtomicUnits as au

class RaSTER(nn.Module):
    """ Range-Separated Transformer with Equivariant Representations

    FID : RASTER

    """

    _graphs_properties: Dict
    dim: int = 176
    """The dimension of the output embedding."""
    nlayers: int = 2
    """The number of message-passing layers."""
    att_dim: int = 16
    """The dimension of the attention heads."""
    scal_heads: int = 16
    """The number of scalar attention heads."""
    tens_heads: int = 4
    """The number of tensor attention heads."""
    lmax: int = 3
    """The maximum angular momentum to consider."""
    normalize_vec: bool = True
    """Whether to normalize the vector features before computing spherical harmonics."""
    att_activation: str = "identity"
    """The activation function to use for the attention coefficients."""
    activation: str = "swish"
    """The activation function to use for the update network."""
    update_hidden: Sequence[int] = ()
    """The hidden layers for the update network."""
    update_bias: bool = True
    """Whether to use bias in the update network."""
    positional_activation: str = "swish"
    """The activation function to use for the positional embedding network."""
    positional_bias: bool = True
    """Whether to use bias in the positional embedding network."""
    switch_before_net: bool = False
    """Whether to apply the switch function to the radial basis before the edge neural network."""
    ignore_parity: bool = False
    """Whether to ignore the parity of the spherical harmonics when constructing the relative positional encoding."""
    additive_positional: bool = False
    """Whether to use additive relative positional encoding. If False, multiplicative relative positional encoding is used."""
    edge_value: bool = False
    """Whether to use edge values in the attention mechanism."""
    layer_normalization: bool = True
    """Whether to use layer normalization of atomic embeddings."""
    graph_key: str = "graph"
    """ The key in the input dictionary that corresponds to the radial graph."""
    embedding_key: str = "embedding"
    """ The key in the output dictionary that corresponds to the embedding."""
    radial_basis: dict = dataclasses.field(
        default_factory=lambda: {"start": 0.8, "basis": "gaussian", "dim": 16}
    )
    """The dictionary of parameters for radial basis functions. See `fennol.models.misc.encodings.RadialBasis`."""
    species_encoding: str | dict = dataclasses.field(
        default_factory=lambda: {"dim": 16, "trainable": True, "encoding": "random"}
    )
    """The dictionary of parameters for species encoding. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    graph_lode: Optional[str] = None
    """The key in the input dictionary that corresponds to the long-range graph."""
    lmax_lode: int = 0
    """The maximum angular momentum for the long-range features."""
    lode_rshort: Optional[float] = None
    """The short-range cutoff for the long-range features."""
    lode_dshort: float = 2.0
    """The width of the short-range cutoff for the long-range features."""
    lode_extra_powers: Sequence[int] = ()
    """The extra powers to include in the long-range features."""
    a_lode: float = -1.0
    """The damping parameter for the long-range features. If negative, the damping is trainable with initial value abs(a_lode)."""
    block_index_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the block index for the MoE network. If None, a normal neural network is used."""
    lode_channels: int = 1
    """The number of channels for the long-range features."""
    switch_cov_start: float = 0.5
    """The start of close-range covalent switch (in units of covalent radii)."""
    switch_cov_end: float = 0.6
    """The end of close-range covalent switch (in units of covalent radii)."""
    normalize_keys: bool = False
    """Whether to normalize queries and keys in the attention mechanism."""
    keep_all_layers: bool = False
    """Whether to return the stacked scalar embeddings from all message-passing layers."""
    
    FID: ClassVar[str] = "RASTER"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]

        ## SETUP LAYER NORMALIZATION
        def _layer_norm(x):
            mu = jnp.mean(x, axis=-1, keepdims=True)
            dx = x - mu
            var = jnp.mean(dx**2, axis=-1, keepdims=True)
            sig = (1.0e-6 + var) ** (-0.5)
            return dx * sig

        if self.layer_normalization:
            layer_norm = _layer_norm
        else:
            layer_norm = lambda x: x
        
        if self.normalize_keys:
            ln_qk = _layer_norm
        else:
            ln_qk = lambda x: x

        ## SPECIES ENCODING
        if isinstance(self.species_encoding, str):
            Zi = inputs[self.species_encoding]
        else:
            Zi = SpeciesEncoding(**self.species_encoding)(species)

        ## INITIALIZE SCALAR FEATURES
        xi = layer_norm(nn.Dense(self.dim, use_bias=False,name="species_linear")(Zi))

        # RADIAL GRAPH
        graph = inputs[self.graph_key]
        distances = graph["distances"]
        switch = graph["switch"]
        edge_src = graph["edge_src"]
        edge_dst = graph["edge_dst"]
        vec = (
            graph["vec"] / graph["distances"][:, None]
            if self.normalize_vec
            else graph["vec"]
        )
        ## CLOSE-RANGE SWITCH
        use_switch_cov = False
        if self.switch_cov_end > 0 and self.switch_cov_start > 0:
            use_switch_cov = True
            assert self.switch_cov_start < self.switch_cov_end, f"switch_cov_start {self.switch_cov_start} must be smaller than switch_cov_end {self.switch_cov_end}"
            assert self.switch_cov_start > 0 and self.switch_cov_end < 1, f"switch_cov_start {self.switch_cov_start} and switch_cov_end {self.switch_cov_end} must be between 0 and 1"
            rc = jnp.array(D3_COV_RADII*au.BOHR)[species]
            rcij = rc[edge_src] + rc[edge_dst]
            rstart = rcij * self.switch_cov_start
            rend = rcij * self.switch_cov_end
            switch_short = (distances >= rend) + 0.5*(1-jnp.cos(jnp.pi*(distances - rstart)/(rend-rstart)))*(distances > rstart)*(distances < rend)
            switch = switch * switch_short

        ## COMPUTE SPHERICAL HARMONICS ON EDGES
        Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(vec)[:,None,:]
        nrep = np.array([2 * l + 1 for l in range(self.lmax + 1)])
        ls = np.arange(self.lmax + 1).repeat(nrep)
            
        parity = jnp.array((-1) ** ls[None,None,:])
        if self.ignore_parity:
            parity = -jnp.ones_like(parity)

        ## INITIALIZE TENSOR FEATURES
        Vi = 0. #jnp.zeros((Zi.shape[0],self.tens_heads, Yij.shape[1]))

        # RADIAL BASIS
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_terms = RadialBasis(
            **{
                **self.radial_basis,
                "end": cutoff,
                "name": f"RadialBasis",
            }
        )(distances)
        if self.switch_before_net:
            radial_terms = radial_terms * switch[:, None]
        elif use_switch_cov:
            radial_terms = radial_terms * switch_short[:, None]

        ## INITIALIZE LODE
        do_lode = self.graph_lode is not None
        if do_lode:
            ## LONG-RANGE GRAPH
            graph_lode = inputs[self.graph_lode]
            switch_lode = graph_lode["switch"][:, None]
            edge_src_lr, edge_dst_lr = graph_lode["edge_src"], graph_lode["edge_dst"]
            r = graph_lode["distances"][:, None]
            rc = self._graphs_properties[self.graph_lode]["cutoff"]

            lmax_lr = self.lmax_lode
            equivariant_lode = lmax_lr > 0
            assert lmax_lr >= 0, f"lmax_lode must be >= 0, got {lmax_lr}"
            assert (
                lmax_lr <= self.lmax
            ), f"lmax_lode must be <= lmax for multipole interaction, got {lmax_lr} > {self.lmax}"
            nrep_lr = np.array([2 * l + 1 for l in range(lmax_lr + 1)], dtype=np.int32)
            if equivariant_lode:
                ls_lr = np.arange(lmax_lr + 1)
            else:
                ls_lr = np.array([0])

            ## PARAMETERS FOR THE LR RADIAL BASIS
            nextra_powers = len(self.lode_extra_powers)
            if nextra_powers > 0:
                ls_lr = np.concatenate([self.lode_extra_powers, ls_lr])

            if self.a_lode > 0:
                a = self.a_lode**2
            else:
                a = (
                    self.param(
                        "a_lr",
                        lambda key: jnp.asarray([-self.a_lode] * ls_lr.shape[0])[
                            None, :
                        ],
                    )
                    ** 2
                )
            rc2a = rc**2 + a
            ls_lr = 0.5 * (ls_lr[None, :] + 1)
            ### minimal radial basis for long range (damped coulomb)
            eij_lr = (
                1.0 / (r**2 + a) ** ls_lr
                - 1.0 / rc2a**ls_lr
                + (r - rc) * rc * (2 * ls_lr) / rc2a ** (ls_lr + 1)
            ) * switch_lode

            if self.lode_rshort is not None:
                rs = self.lode_rshort
                d = self.lode_dshort
                switch_short = 0.5 * (1 - jnp.cos(jnp.pi * (r - rs) / d)) * (r > rs) * (
                    r < rs + d
                ) + (r >= rs + d)
                eij_lr = eij_lr * switch_short

            dim_lr = 1
            if nextra_powers > 0:
                eij_lr_extra = eij_lr[:, :nextra_powers]
                eij_lr = eij_lr[:, nextra_powers:]
                dim_lr += nextra_powers

            if equivariant_lode:
                ## SPHERICAL HARMONICS ON LONG-RANGE GRAPH
                eij_lr = eij_lr.repeat(nrep_lr, axis=-1)
                Yij_lr = generate_spherical_harmonics(lmax=lmax_lr, normalize=False)(
                    graph_lode["vec"] / r
                )
                dim_lr += lmax_lr
                eij_lr = eij_lr * Yij_lr
                del Yij_lr
        

        if self.keep_all_layers:
            fis = []

        ### START MESSAGE PASSING ITERATIONS
        for layer in range(self.nlayers):
            ## GATHER SCALAR EDGE FEATURES
            u = [radial_terms]
            if layer > 0:
                ## edge-tensor contraction
                xij2 = (Vi[edge_dst] + (parity* Vi)[edge_src]) * Yij
                for l in range(self.lmax + 1):
                    u.append((xij2[:,:, l**2 : (l + 1) ** 2]).sum(axis=-1))
            ur = jnp.concatenate(u, axis=-1)

            ## BUILD RELATIVE POSITIONAL ENCODING
            if self.edge_value:
                nout = 2
            else:
                nout = 1
            w = FullyConnectedNet(
                [2 * self.att_dim, nout*self.att_dim],
                activation=self.positional_activation,
                use_bias=self.positional_bias,
                name=f"positional_encoding_{layer}",
            )(ur).reshape(radial_terms.shape[0],nout, self.att_dim)
            if self.edge_value:
                w,vij = jnp.split(w, 2, axis=1)

            nls = self.lmax + 1 if layer == 0 else 2 * (self.lmax + 1)


            ## QUERY, KEY, VALUE
            q = ln_qk(nn.Dense((self.scal_heads + nls*self.tens_heads) * self.att_dim, use_bias=False,name=f"queries_{layer}")(
                xi
            ).reshape(xi.shape[0], self.scal_heads + nls*self.tens_heads, self.att_dim))
            k = nn.Dense((self.scal_heads + nls*self.tens_heads) * self.att_dim, use_bias=False, name=f"keys_{layer}")(
                xi
            ).reshape(xi.shape[0], self.scal_heads + nls*self.tens_heads, self.att_dim)

            v = nn.Dense(self.scal_heads * self.att_dim, use_bias=False, name=f"values_{layer}")(xi).reshape(
                xi.shape[0], self.scal_heads, self.att_dim
            )

            ## ATTENTION COEFFICIENTS
            if self.additive_positional:
                wk = ln_qk(w + k[edge_dst])
            else:
                wk = ln_qk(w * k[edge_dst])

            act = activation_from_str(self.att_activation)
            aij = (
                act((q[edge_src] * wk).sum(axis=-1) / (self.att_dim**0.5))
                * switch[:, None]
            )

            aijl = aij[:, : self.tens_heads*(self.lmax + 1)].reshape(-1,self.tens_heads,self.lmax+1).repeat(nrep, axis=-1)
            if layer > 0:
                aijl1 = aij[:, self.tens_heads*(self.lmax + 1) : self.tens_heads*nls].reshape(-1,self.tens_heads,self.lmax+1).repeat(nrep, axis=-1)
            aij = aij[:, self.tens_heads*nls:, None]

            if self.edge_value:
                ## EDGE VALUES
                if self.additive_positional:
                    vij = vij + v[edge_dst]
                else:
                    vij = vij * v[edge_dst]
            else:
                ## MOVE DEST VALUES TO EDGE
                vij = v[edge_dst]

            ## SCALAR ATTENDED FEATURES
            vai = jax.ops.segment_sum(
                aij * vij,
                edge_src,
                num_segments=xi.shape[0],
            )
            vai = vai.reshape(xi.shape[0], -1)

            ### TENSOR ATTENDED FEATURES
            uij = aijl * Yij
            if layer > 0:
                uij = uij + aijl1 * Vi[edge_dst]
            Vi = Vi + jax.ops.segment_sum(uij, edge_src, num_segments=Zi.shape[0])

            ## SELF SCALAR FEATURES
            si = nn.Dense(self.att_dim, use_bias=False, name=f"self_values_{layer}")(xi)

            components = [si, vai]

            ### CONTRACT TENSOR FEATURES TO BUILD INVARIANTS
            if self.tens_heads == 1:
                Vi2 = Vi**2
            else:
                Vi2 = Vi * ChannelMixing(self.lmax, self.tens_heads, name=f"extract_mixing_{layer}")(Vi)
            for l in range(self.lmax + 1):
                norm = 1.0 / (2 * l + 1)
                components.append(
                    (Vi2[:,:, l**2 : (l + 1) ** 2]).sum(axis=-1) * norm
                )

            ### LODE (~ LONG-RANGE ATTENTION)
            if do_lode and layer == self.nlayers - 1:
                assert self.lode_channels <= self.tens_heads
                zj = nn.Dense(self.lode_channels*dim_lr, use_bias=False, name=f"lode_values_{layer}")(xi).reshape(
                    xi.shape[0], self.lode_channels, dim_lr
                )
                if nextra_powers > 0:
                    zj_extra = zj[:,:, :nextra_powers]
                    zj = zj[:, :, nextra_powers:]
                    xi_lr_extra = jax.ops.segment_sum(
                        eij_lr_extra[:,None,:] * zj_extra[edge_dst_lr],
                        edge_src_lr,
                        species.shape[0],
                    ).reshape(species.shape[0],-1)
                    components.append(xi_lr_extra)
                if equivariant_lode:
                    zj = zj.repeat(nrep_lr, axis=-1)
                Vi_lr = jax.ops.segment_sum(
                    eij_lr[:,None,:] * zj[edge_dst_lr], edge_src_lr, species.shape[0]
                )
                components.append(Vi_lr[:,: , 0])
                if equivariant_lode:
                    Mi_lr = Vi[:,:self.lode_channels, : (lmax_lr + 1) ** 2] * Vi_lr
                    for l in range(1, lmax_lr + 1):
                        norm = 1.0 / (2 * l + 1)
                        components.append(
                            Mi_lr[:, :,l**2 : (l + 1) ** 2].sum(axis=-1)
                            * norm
                        )

            ### CONCATENATE UPDATE COMPONENTS
            components = jnp.concatenate(components, axis=-1)
            ### COMPUTE UPDATE
            if self.block_index_key is not None:
                ## MoE neural network from block index
                block_index = inputs[self.block_index_key]
                updi = BlockIndexNet(
                        output_dim=self.dim + self.tens_heads*(self.lmax + 1),
                        hidden_neurons=self.update_hidden,
                        activation=self.activation,
                        use_bias=self.update_bias,
                        name=f"update_net_{layer}",
                    )((species,components, block_index))
            else:
                updi = FullyConnectedNet(
                        [*self.update_hidden, self.dim + self.tens_heads*(self.lmax + 1)],
                        activation=self.activation,
                        use_bias=self.update_bias,
                        name=f"update_net_{layer}",
                    )(components)
                
            ## UPDATE ATOM FEATURES
            xi = layer_norm(xi + updi[:,:self.dim])
            Vi = Vi * (1 + updi[:,self.dim:]).reshape(-1,self.tens_heads,self.lmax+1).repeat(nrep, axis=-1)
            if self.tens_heads > 1:
                Vi = ChannelMixing(self.lmax, self.tens_heads,name=f"update_mixing_{layer}")(Vi)

            if self.keep_all_layers:
                ## STORE ALL LAYERS
                fis.append(xi)


        output = {**inputs, self.embedding_key: xi, self.embedding_key + "_tensor": Vi}
        if self.keep_all_layers:
            output[self.embedding_key+'_layers'] = jnp.stack(fis,axis=1)
        return output


import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils.activations import activation_from_str, tssr3
from ..misc.nets import FullyConnectedNet


class PAINNEmbedding(nn.Module):
    """polarizable atom interaction neural network
    
    ### Reference
    K.T. Schütt. P.-J. Kindermans, H. E. Sauceda, S. Chmiela, A. Tkatchenko, K.-R. Müller. SchNet - a deep learning architecture for molecules and materials. The Journal of Chemical Physics 148(24), 241722 (2018) 
    https://doi.org/10.1063/1.5019779

    """

    _graphs_properties: Dict
    dim: int = 128
    """ The dimension of the embedding. """
    nlayers: int = 3
    """ The number of interaction layers. """
    nchannels: Optional[int] = None
    """ The number of equivariant channels. If None, it is set to dim. """
    message_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """ The hidden layers for the message network."""
    update_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """ The hidden layers for the update network."""
    activation: Union[Callable, str] = "silu"
    """ The activation function."""
    graph_key: str = "graph"
    """ The key for the graph input."""
    embedding_key: str = "embedding"
    """ The key for the embedding output."""
    tensor_embedding_key: str = "embedding_vectors"
    """ The key for the tensor embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """ The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """ The radial basis function parameters. See `fennol.models.misc.encodings.RadialBasis`."""
    keep_all_layers: bool = False
    """ Whether to keep the embedding from each layer in the output."""

    FID: ClassVar[str] = "PAINN"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        cutoff = self._graphs_properties[self.graph_key]["cutoff"]

        onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
            species
        )
        xi = nn.Dense(self.dim, name="species_linear", use_bias=True)(onehot)

        nchannels = self.nchannels if self.nchannels is not None else self.dim

        distances = graph["distances"]
        switch = graph["switch"][:, None]
        dirij = (graph["vec"] / distances[:, None])[:, :,None]
        Vi = jnp.zeros((xi.shape[0], 3, nchannels), dtype=xi.dtype)

        radial_basis = RadialBasis(
            **{
                **self.radial_basis,
                "end": cutoff,
                "name": f"RadialBasis",
            }
        )(distances)

        if self.keep_all_layers:
            xis = []
        for layer in range(self.nlayers):
            # compute messages
            phi = FullyConnectedNet(
                [*self.message_hidden, self.dim + 2 * nchannels],
                activation=self.activation,
                name=f"message_{layer}",
                use_bias=True,
            )(xi)
            w = (
                nn.Dense(
                    self.dim + 2 * nchannels,
                    name=f"radial_linear_{layer}",
                    use_bias=True,
                )(radial_basis)
                * switch
            )
            dxij, hvv, hvs = jnp.split(
                phi[edge_dst] * w, [self.dim, self.dim + nchannels], axis=-1
            )

            dvij = dirij * hvs[:, None,:]
            if layer > 0:
                dvij = dvij + Vi[edge_dst] * hvv[:, None,:]

            # aggregate messages
            v_message = Vi + jax.ops.segment_sum(dvij, edge_src, Vi.shape[0])
            x_message = xi + jax.ops.segment_sum(dxij, edge_src, xi.shape[0])

            # update
            u,v = jnp.split(
                nn.Dense(
                    2 * self.nchannels,
                    use_bias=False,
                    name=f"UV_{layer}",
                )(v_message),
                2,
                axis=-1,
            )

            scals = (u * v).sum(axis=1)
            norms = tssr3((v**2).sum(axis=1))

            A = FullyConnectedNet(
                [*self.update_hidden, self.dim + 2 * nchannels],
                activation=self.activation,
                name=f"update_{layer}",
                use_bias=True,
            )(jnp.concatenate((x_message, norms), axis=-1))

            ass, asv, avv = jnp.split(
                A,
                [self.dim, self.dim + nchannels],
                axis=-1,
            )

            Vi = Vi + u * avv[:, None,:]
            if self.dim != nchannels:
                dxi = nn.Dense(self.dim, name=f"resize_{layer}", use_bias=False)(
                    scals * asv
                )
            else:
                dxi = scals * asv

            xi = xi + ass + dxi

            if self.keep_all_layers:
                xis.append(xi)

        output = {
            **inputs,
            self.embedding_key: xi,
            self.tensor_embedding_key: Vi,
        }
        if self.keep_all_layers:
            output[self.embedding_key + "_layers"] = jnp.stack(xis, axis=1)
        return output



#!/usr/bin/env python3
"""Polarization model for FENNOL.

Created by C. Cattin 2024
"""

import flax.linen as nn
import jax
import jax.numpy as jnp
from typing import Optional, ClassVar

from fennol.utils import AtomicUnits as Au


class Polarization(nn.Module):
    """Polarization model with Thole damping scheme.
    
    FID: POLARIZATION
    """

    energy_key: Optional[str] = None
    """Key of the energy in the outputs."""
    graph_key: str = 'graph'
    """Key of the graph in the inputs."""
    polarizability_key: str = 'polarizability'
    """Key of the polarizability in the inputs."""
    electric_field_key: str = 'electric_field'
    """Key of the electric field in the inputs."""
    induced_dipoles_key: str = 'induced_dipoles'
    """Key of the induced dipole in the outputs."""
    damping_param_mutual: float = 0.39
    """Damping parameter for mutual polarization."""
    neglect_mutual: bool = False
    """Neglect the mutual polarization term (like in iAMOEBA)."""
    _energy_unit: str = 'Ha'
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str] = 'POLARIZATION'

    @nn.compact
    def __call__(self, inputs):
        """Forward pass of the polarization model.

        Parameters
        ----------
        inputs : dict
            Input dictionary containing all the info about the system.
            This dictionary is given from the FENNIX class.
        """
        # Species
        species = inputs['species']
        graph = inputs[self.graph_key]
        # Graph
        edge_src, edge_dst = graph['edge_src'], graph['edge_dst']
        # Distances and vector between each pair of atoms in atomic units
        distances = graph['distances']
        rij = distances / Au.BOHR
        vec_ij = graph['vec'] / Au.BOHR
        # Polarizability
        polarizability = (
            inputs[self.polarizability_key] / Au.BOHR**3
        )

        pol_src = polarizability[edge_src]
        pol_dst = polarizability[edge_dst]
        alpha_ij = pol_dst * pol_src

        # The output is a dictionary with the polarization energy
        output = {}

        ######################
        # Interaction matrix #
        ######################
        # Diagonal terms
        tii = 1 / polarizability[:, None]
        

        ##################
        # Electric field #
        ##################
        electric_field = inputs[self.electric_field_key]

        ###############################
        # Electric point dipole moment#
        ###############################
        if self.neglect_mutual:
            electric_field = electric_field.reshape(-1, 3)
            mu = polarizability[:,None]*electric_field
            mu_ = jax.lax.stop_gradient(mu)
            tmu = tii*mu_
            
        else:
            # Effective distance
            uij = rij / alpha_ij ** (1 / 6)
            # Damping terms
            exp = jnp.exp(-self.damping_param_mutual * uij**3)
            lambda_3 = 1 - exp
            lambda_5 = 1 - (1 + self.damping_param_mutual * uij**3) * exp
            # Off-diagonal terms
            tij = (
                3 * lambda_5[:, None, None]
                    * vec_ij[:, :, None] * vec_ij[:, None, :]
                    / rij[:, None, None]**5
                - jnp.eye(3)[None, :, :] * lambda_3[:, None, None]
                    / rij[:, None, None]**3
            )

            def matvec(mui):
                """Compute the matrix vector product of T and mu."""
                mui = mui.reshape(-1, 3)
                tmu_self = tii * mui
                tmu_pair = jnp.einsum("jab,jb->ja", tij, mui[edge_dst])
                tmu = (
                    jax.ops.segment_sum(tmu_pair, edge_src, species.shape[0])
                    + tmu_self
                )
                return tmu.flatten()
            mu = jax.scipy.sparse.linalg.cg(matvec, electric_field)[0]
            mu_ = jax.lax.stop_gradient(mu)

            # Matrix vector product
            tmu = matvec(mu_)

        # Polarization energy
        pol_energy = (
            (0.5 * tmu - electric_field) * mu_
        ).reshape(-1, 3).sum(axis=1)

        # Output
        output[self.electric_field_key] = electric_field.reshape(-1, 3)
        output[self.induced_dipoles_key] = mu.reshape(-1, 3) * Au.BOHR
        energy_key = (
            self.energy_key if self.energy_key is not None else self.name
        )
        energy_unit = Au.get_multiplier(self._energy_unit)
        output[energy_key] = pol_energy*energy_unit
        output['tmu'] = tmu.reshape(-1, 3)

        return {**inputs, **output}


if __name__ == "__main__":
    pass



import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils.activations import activation_from_str
from ..misc.nets import FullyConnectedNet


class NewtonNetEmbedding(nn.Module):
    """ Newtonian message passing network

    ### Reference
    Haghighatlari et al., NewtonNet: a Newtonian message passing network for deep learning of interatomic potentials and forces
    https://doi.org/10.1039/D2DD00008C

    """
    
    _graphs_properties: Dict
    dim: int = 128
    """The dimension of the embedding."""
    nlayers: int = 3
    """The number of interaction layers."""
    nchannels: Optional[int] = None
    """The number of vector channels. If None, it is set to dim."""
    embedding_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """The hidden layers for the embedding networks."""
    latent_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [128])
    """The hidden layers for the latent update network."""
    activation: Union[Callable, str] = "silu"
    """The activation function."""
    graph_key: str = "graph"
    """The key for the graph input."""
    embedding_key: str = "embedding"
    """The key for the output embedding."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`."""
    keep_all_layers: bool = False
    """Whether to keep embeddings from each layer in the output."""

    FID: ClassVar[str] = "NEWTONNET"


    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        cutoff = self._graphs_properties[self.graph_key]["cutoff"]

        onehot = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(
            species
        )
        xi = nn.Dense(self.dim, name="species_linear", use_bias=True)(onehot)

        nchannels = self.nchannels if self.nchannels is not None else self.dim

        distances = graph["distances"]
        switch = graph["switch"][:, None]
        dirij = graph["vec"] / distances[:, None] * switch

        radial_basis = RadialBasis(
            **{
                **self.radial_basis,
                "end": cutoff,
                "name": f"RadialBasis",
            }
        )(distances)

        if self.keep_all_layers:
            xis = []
        for layer in range(self.nlayers):
            ai = FullyConnectedNet(
                [*self.embedding_hidden, self.dim],
                activation=self.activation,
                name=f"phi_a_{layer}",
                use_bias=True,
            )(xi)
            Dij = nn.Dense(self.dim, name=f"radial_linear_{layer}", use_bias=True)(
                radial_basis
            )
            mij = ai[edge_src] * ai[edge_dst] * Dij * switch

            mi = jax.ops.segment_sum(mij, edge_src, xi.shape[0])
            xi = xi + mi

            Fij = (
                FullyConnectedNet(
                    [*self.embedding_hidden, 1],
                    activation=self.activation,
                    name=f"phi_F_{layer}",
                    use_bias=True,
                )(mij)
                * dirij
            )

            fij = (
                FullyConnectedNet(
                    [*self.embedding_hidden, nchannels],
                    activation=self.activation,
                    name=f"phi_f_{layer}",
                    use_bias=True,
                )(mij)[:, :, None]
                * Fij[:, None, :]
            )

            if layer == 0:
                fi = jax.ops.segment_sum(fij, edge_src, xi.shape[0])
            else:
                fi = fi + jax.ops.segment_sum(fij, edge_src, xi.shape[0])

            deltai = (
                FullyConnectedNet(
                    [*self.embedding_hidden, nchannels],
                    activation=self.activation,
                    name=f"phi_R_{layer}",
                    use_bias=True,
                )(xi)[:, :, None]
                * fi
            )
            if layer == 0:
                di = deltai
            else:
                phi_rij = FullyConnectedNet(
                    [*self.embedding_hidden, nchannels],
                    activation=self.activation,
                    name=f"phi_r_{layer}",
                    use_bias=True,
                )(mij)
                
                phi_r = jax.ops.segment_sum(phi_rij * switch, edge_src, xi.shape[0])
                di = phi_r[:, :, None] * di + deltai

            scal = jnp.sum(fi * di, axis=-1)
            dui = (
                -FullyConnectedNet(
                    [*self.latent_hidden, nchannels],
                    activation=self.activation,
                    name=f"phi_u_{layer}",
                    use_bias=True,
                )(xi)
                * scal
            )

            if nchannels != self.dim:
                dui = nn.Dense(self.dim, name=f"reshape_{layer}", use_bias=False)(dui)

            xi = xi + dui
            if self.keep_all_layers:
                xis.append(xi)

        output = {
            **inputs,
            self.embedding_key: xi,
        }
        if self.keep_all_layers:
            output[self.embedding_key + "_layers"] = jnp.stack(xis, axis=1)
        return output



import flax.linen as nn
import jax
import jax.numpy as jnp
from typing import Callable,Union

def scaled_orthogonal(
    scale=1.0, mode="fan_avg", in_axis=-2, out_axis=-1, dtype=jnp.float_
):
    assert mode in ["fan_in", "fan_out", "fan_avg"]
    init_ortho = nn.initializers.orthogonal(
        scale=scale, column_axis=out_axis, dtype=dtype
    )
    if mode == "fan_in":

        def init(key, shape, dtype=jnp.float32):
            return init_ortho(key, shape, dtype=dtype) * (shape[in_axis] ** -0.5)

    elif mode == "fan_out":

        def init(key, shape, dtype=jnp.float32):
            return init_ortho(key, shape, dtype=dtype) * (shape[out_axis] ** -0.5)

    else:

        def init(key, shape, dtype=jnp.float32):
            return init_ortho(key, shape, dtype=dtype) * (
                (shape[in_axis] + shape[out_axis]) ** -0.5
            )

    return init


def initializer_from_str(name: Union[str,Callable])->Callable:
    if callable(name):
        return name
    if not isinstance(name, str):
        raise ValueError(f"Invalid initializer {name}")
    return eval(
        name,
        {"__builtins__": None},
        {
            **nn.initializers.__dict__,
            "scaled_orthogonal": scaled_orthogonal,
        },
    )



import jax
import jax.numpy as jnp
import flax.linen as nn
import dataclasses
import numpy as np
from typing import Dict, Union, Callable, Sequence, Optional, Tuple, ClassVar

from ..misc.encodings import SpeciesEncoding, RadialBasis, positional_encoding
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ...utils.activations import activation_from_str
from ...utils.initializers import initializer_from_str
from ..misc.nets import FullyConnectedNet, BlockIndexNet
from ..misc.e3 import ChannelMixing, ChannelMixingE3, FilteredTensorProduct
from ...utils.periodic_table import D3_COV_RADII, D3_VDW_RADII, VALENCE_ELECTRONS
from ...utils import AtomicUnits as au

class CRATEmbedding(nn.Module):
    """Configurable Resources ATomic Environment

    FID : CRATE

    This class represents the CRATE (Configurable Resources ATomic Environment) embedding model.
    It is used to encode atomic environments using multiple sources of information
      (radial, angular, E(3), message-passing, LODE, etc...)
    """

    _graphs_properties: Dict

    dim: int = 256
    """The size of the embedding vectors."""
    nlayers: int = 2
    """The number of interaction layers in the model."""
    keep_all_layers: bool = False
    """Whether to output all layers."""

    dim_src: int = 64
    """The size of the source embedding vectors."""
    dim_dst: int = 32
    """The size of the destination embedding vectors."""

    angle_style: str = "fourier"
    """The style of angle representation."""
    dim_angle: int = 8
    """The size of the pairwise vectors use for triplet combinations."""
    nmax_angle: int = 4
    """The dimension of the angle representation (minus one)."""
    zeta: float = 14.1
    """The zeta parameter for the model ANI angular representation."""
    angle_combine_pairs: bool = True
    """Whether to combine angle pairs instead of average distance embedding like in ANI."""

    message_passing: bool = True
    """Whether to use message passing in the model."""
    att_dim: int = 1
    """The hidden size for the attention mechanism (only used when message-passing is disabled)."""

    lmax: int = 0
    """The maximum order of spherical tensors."""
    nchannels_l: int = 16
    """The number of channels for spherical tensors."""
    n_tp: int = 1
    """The number of tensor products performed at each layer."""
    ignore_irreps_parity: bool = False
    """Whether to ignore the parity of the irreps in the tensor product."""
    edge_tp: bool = False
    """Whether to perform a tensor product on edges before sending messages."""
    resolve_wij_l: bool = False
    """Equivariant message weights are l-dependent."""

    species_init: bool = False
    """Whether to initialize the embedding using the species encoding."""
    mixing_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
    """The hidden layer sizes for the mixing network."""
    pair_mixing_hidden: Sequence[int] = dataclasses.field(default_factory=lambda: [])
    """The hidden layer sizes for the pair mixing network."""
    activation: Union[Callable, str] = "silu"
    """The activation function for the mixing network."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization function for Dense operations."""
    activation_mixing: Union[Callable, str] = "tssr3"
    """The activation function applied after mixing."""
    layer_normalization: bool = False
    """Whether to apply layer normalization after each layer."""
    use_bias: bool = True
    """Whether to use bias in the Dense operations."""

    graph_key: str = "graph"
    """The key for the graph data in the inputs dictionary."""
    graph_angle_key: Optional[str] = None
    """The key for the angle graph data in the inputs dictionary."""
    embedding_key: Optional[str] = None
    """The key for the embedding data in the output dictionary."""
    pair_embedding_key: Optional[str] = None
    """The key for the pair embedding data in the output dictionary."""

    species_encoding: Union[dict, str] = dataclasses.field(default_factory=dict)
    """If `str`, it is the key in the inputs dictionary that contains species encodings. Else, it is the dictionary of parameters for species encoding. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The dictionary of parameters for radial basis functions. See `fennol.models.misc.encodings.RadialBasis`."""
    radial_basis_angle: Optional[dict] = None
    """The dictionary of parameters for radial basis functions for angle embedding. 
        If None, the radial basis for angles is the same as the radial basis for distances."""

    graph_lode: Optional[str] = None
    """The key for the lode graph data in the inputs dictionary."""
    lode_channels: Union[int, Sequence[int]] = 8
    """The number of channels for lode."""
    lmax_lode: int = 0
    """The maximum order of spherical tensors for lode."""
    a_lode: float = -1.
    """The cutoff for the lode graph. If negative, the value is trainable with starting value -a_lode."""
    lode_resolve_l: bool = True
    """Whether to resolve the lode channels by l."""
    lode_multipole_interaction: bool = True
    """Whether to interact with the multipole moments of the lode graph."""
    lode_direct_multipoles: bool = True
    """Whether to directly use the first local equivariants to interact with long-range equivariants. If false, local equivariants are mixed before interaction."""
    lode_equi_full_combine: bool = False
    lode_normalize_l: bool = False
    lode_use_field_norm: bool = True
    lode_rshort: Optional[float] = None
    lode_dshort: float = 0.5
    lode_extra_powers: Sequence[int] = ()
    

    charge_embedding: bool = False
    """Whether to include charge embedding."""
    total_charge_key: str = "total_charge"
    """The key for the total charge data in the inputs dictionary."""

    block_index_key: Optional[str] = None
    """The key for the block index. If provided, will use a BLOCK_INDEX_NET as a mixing network."""

    FID: ClassVar[str] = "CRATE"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"
        reduce_memory =  "reduce_memory" in inputs.get("flags", {})

        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )

        actmix = activation_from_str(self.activation_mixing)

        ##################################################
        graph = inputs[self.graph_key]
        use_angles = self.graph_angle_key is not None
        if use_angles:
            graph_angle = inputs[self.graph_angle_key]

            # Check that the graph_angle is a subgraph of graph
            correct_graph = (
                self.graph_angle_key == self.graph_key
                or self._graphs_properties[self.graph_angle_key]["parent_graph"]
                == self.graph_key
            )
            assert (
                correct_graph
            ), f"graph_angle_key={self.graph_angle_key} must be a subgraph of graph_key={self.graph_key}"
            assert (
                "angles" in graph_angle
            ), f"Graph {self.graph_angle_key} must contain angles"
            # check if graph_angle is a filtered graph
            filtered = "parent_graph" in self._graphs_properties[self.graph_angle_key]
            if filtered:
                filter_indices = graph_angle["filter_indices"]

        ##################################################
        ### SPECIES ENCODING ###
        if isinstance(self.species_encoding, str):
            zi = inputs[self.species_encoding]
        else:
            zi = SpeciesEncoding(
                **self.species_encoding, name="SpeciesEncoding"
            )(species)

        
        if self.layer_normalization:
            def layer_norm(x):
                mu = jnp.mean(x,axis=-1,keepdims=True)
                dx = x-mu
                var = jnp.mean(dx**2,axis=-1,keepdims=True)
                sig = (1.e-6 + var)**(-0.5)
                return dx*sig
        else:
            layer_norm = lambda x:x
                

        if self.charge_embedding:
            xi, qi = jnp.split(
                nn.Dense(self.dim + 1, use_bias=False, name="ChargeEncoding")(zi),
                [self.dim],
                axis=-1,
            )
            batch_index = inputs["batch_index"]
            natoms = inputs["natoms"]
            nsys = natoms.shape[0]
            Zi = jnp.asarray(VALENCE_ELECTRONS)[species]
            Ntot = jax.ops.segment_sum(Zi, batch_index, nsys) - inputs.get(
                self.total_charge_key, jnp.zeros(nsys)
            )
            ai = jax.nn.softplus(qi.squeeze(-1))
            A = jax.ops.segment_sum(ai, batch_index, nsys)
            Ni = ai * (Ntot / A)[batch_index]
            charge_embedding = positional_encoding(Ni, self.dim)
            xi = layer_norm(xi + charge_embedding)
        elif self.species_init:
            xi = layer_norm(nn.Dense(self.dim, use_bias=False, name="SpeciesInit")(zi))
        else:
            xi = zi

        ##################################################
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        distances = graph["distances"]
        switch = graph["switch"][:, None]

        ### COMPUTE RADIAL BASIS ###
        radial_basis = RadialBasis(
            **{
                **self.radial_basis,
                "end": cutoff,
                "name": f"RadialBasis",
            }
        )(distances)

        do_lode = self.graph_lode is not None
        if do_lode:
            graph_lode = inputs[self.graph_lode]
            switch_lode = graph_lode["switch"][:, None]

            edge_src_lr, edge_dst_lr = graph_lode["edge_src"], graph_lode["edge_dst"]
            r = graph_lode["distances"][:, None]
            rc = self._graphs_properties[self.graph_lode]["cutoff"]
            
            lmax_lr = self.lmax_lode
            equivariant_lode = lmax_lr > 0
            assert lmax_lr >=0, f"lmax_lode must be >= 0, got {lmax_lr}"
            if self.lode_multipole_interaction:
                assert lmax_lr <= self.lmax, f"lmax_lode must be <= lmax for multipole interaction, got {lmax_lr} > {self.lmax}"
            nrep_lr = np.array([2 * l + 1 for l in range(lmax_lr + 1)], dtype=np.int32)
            if self.lode_resolve_l and equivariant_lode:
                ls_lr = np.arange(lmax_lr + 1)
            else:
                ls_lr = np.array([0])

            nextra_powers = len(self.lode_extra_powers)
            if nextra_powers > 0:
                ls_lr = np.concatenate([self.lode_extra_powers,ls_lr])

            if self.a_lode > 0:
                a = self.a_lode**2
            else:
                a = (
                    self.param(
                        "a_lr",
                        lambda key: jnp.asarray([-self.a_lode] * ls_lr.shape[0])[None, :],
                    )
                    ** 2
                )
            rc2a = rc**2 + a
            ls_lr = 0.5 * (ls_lr[None, :] + 1)
            ### minimal radial basis for long range (damped coulomb)
            eij_lr = (
                1.0 / (r**2 + a) ** ls_lr
                - 1.0 / rc2a**ls_lr
                + (r - rc) * rc * (2 * ls_lr) / rc2a ** (ls_lr + 1)
            ) * switch_lode

            if self.lode_rshort is not None:
                rs = self.lode_rshort
                d = self.lode_dshort
                switch_short = 0.5 * (1 - jnp.cos(jnp.pi * (r - rs) / d)) * (r > rs) * (
                    r < rs + d
                ) + (r >= rs + d)
                eij_lr = eij_lr * switch_short
            
            if nextra_powers>0:
                eij_lr_extra = eij_lr[:,:nextra_powers]
                eij_lr = eij_lr[:,nextra_powers:]


            # dim_lr = self.nchannels_lode
            nchannels_lode = (
                [self.lode_channels] * self.nlayers
                if isinstance(self.lode_channels, int)
                else self.lode_channels
            )
            dim_lr = nchannels_lode
            
            if equivariant_lode:
                if self.lode_resolve_l:
                    eij_lr = eij_lr.repeat(nrep_lr, axis=-1)
                Yij = generate_spherical_harmonics(lmax=lmax_lr, normalize=False)(
                    graph_lode["vec"] / r
                )
                eij_lr = (eij_lr * Yij)[:, None, :]
                dim_lr = [d * (lmax_lr + 1) for d in dim_lr]
            
            if nextra_powers > 0:
                eij_lr_extra = eij_lr_extra[:,None,:]
                extra_dims = [nextra_powers*d for d in nchannels_lode]
                dim_lr = [d + ed for d,ed in zip(dim_lr,extra_dims)]
            

        ##################################################
        ### GET ANGLES ###
        if use_angles:
            angles = graph_angle["angles"][:, None]
            angle_src, angle_dst = graph_angle["angle_src"], graph_angle["angle_dst"]
            switch_angles = graph_angle["switch"][:, None]
            central_atom = graph_angle["central_atom"]

            if not self.angle_combine_pairs:
                assert (
                    self.radial_basis_angle is not None
                ), "radial_basis_angle must be specified if angle_combine_pairs=False"

            ### COMPUTE RADIAL BASIS FOR ANGLES ###
            if self.radial_basis_angle is not None:
                dangles = graph_angle["distances"]
                swang = switch_angles
                if not self.angle_combine_pairs:
                    dangles = 0.5 * (dangles[angle_src] + dangles[angle_dst])
                    swang = switch_angles[angle_src] * switch_angles[angle_dst]
                radial_basis_angle = (
                    RadialBasis(
                        **{
                            **self.radial_basis_angle,
                            "end": self._graphs_properties[self.graph_angle_key][
                                "cutoff"
                            ],
                            "name": f"RadialBasisAngle",
                        }
                    )(dangles)
                    * swang
                )

            else:
                if filtered:
                    radial_basis_angle = radial_basis[filter_indices] * switch_angles
                else:
                    radial_basis_angle = radial_basis * switch

        radial_basis = radial_basis * switch

        # # add covalent indicator
        # rc = jnp.asarray([d/au.BOHR for d in D3_COV_RADII])[species]
        # rcij = rc[edge_src] + rc[edge_dst]
        # fact = graph["switch"]*(2*distances/rcij)*jnp.exp(-0.5 * ((distances - rcij)/(0.1*rcij)) ** 2)
        # radial_basis = jnp.concatenate([radial_basis,fact[:,None]],axis=-1)
        # if use_angles:
        #     rcij = rc[graph_angle["edge_src"]] + rc[graph_angle["edge_dst"]]
        #     dangles = graph_angle["distances"]
        #     fact = graph_angle["switch"]*((2*dangles/rcij))*jnp.exp(-0.5 * ((dangles - rcij)/(0.1*rcij))**2)
        #     radial_basis_angle = jnp.concatenate([radial_basis_angle,fact[:,None]],axis=-1)


        ##################################################
        if self.lmax > 0:
            Yij = generate_spherical_harmonics(lmax=self.lmax, normalize=False)(
                graph["vec"] / graph["distances"][:, None]
            )[:, None, :]
            Yij = jnp.broadcast_to(Yij, (Yij.shape[0], self.nchannels_l, Yij.shape[2]))
            nrep_l = np.array([2 * l + 1 for l in range(self.lmax + 1)], dtype=np.int32)
            # ls = [0]
            # for l in range(1, self.lmax + 1):
            #    ls = ls + [l] * (2 * l + 1)
            #ls = jnp.asarray(np.array(ls)[None, :], dtype=distances.dtype)
            #lcut = (0.5 + 0.5 * jnp.cos((np.pi / cutoff) * distances[:, #None])) ** (
            #    ls + 1
            #)
            # lcut = jnp.where(graph["edge_mask"][:, None], lcut, 0.0)
            # rijl1 = (lcut * distances[:, None] ** ls)[:, None, :]

        ##################################################
        if use_angles:
            ### ANGULAR BASIS ###
            if self.angle_style == "fourier":
                # build fourier series for angles
                nangles = self.param(
                    f"nangles",
                    lambda key, dim: jnp.arange(dim, dtype=distances.dtype)[None, :],
                    self.nmax_angle + 1,
                )

                phi = self.param(
                    f"phi",
                    lambda key, dim: jnp.zeros((1, dim), dtype=distances.dtype),
                    self.nmax_angle + 1,
                )
                xa = jnp.cos(nangles * angles + phi)
            elif self.angle_style == "fourier_full":
                # build fourier series for angles including sin terms
                nangles = self.param(
                    f"nangles",
                    lambda key, dim: jnp.arange(dim, dtype=distances.dtype)[None, :],
                    self.nmax_angle + 1,
                )

                phi = self.param(
                    f"phi",
                    lambda key, dim: jnp.zeros((1, dim), dtype=distances.dtype),
                    2 * self.nmax_angle + 1,
                )
                xac = jnp.cos(nangles * angles + phi[:, : self.nmax_angle + 1])
                xas = jnp.sin(nangles[:, 1:] * angles + phi[:, self.nmax_angle + 1 :])
                xa = jnp.concatenate([xac, xas], axis=-1)
            elif self.angle_style == "ani":
                # ANI-style angle embedding
                angle_start = np.pi / (2 * (self.nmax_angle + 1))
                shiftZ = self.param(
                    f"shiftZ",
                    lambda key, dim: jnp.asarray(
                        (np.linspace(0, np.pi, dim + 1) + angle_start)[None, :-1],
                        dtype=distances.dtype,
                    ),
                    self.nmax_angle + 1,
                )
                zeta = self.param(
                    f"zeta",
                    lambda key: jnp.asarray(self.zeta, dtype=distances.dtype),
                )
                xa = (0.5 + 0.5 * jnp.cos(angles - shiftZ)) ** zeta
            else:
                raise ValueError(f"Unknown angle style {self.angle_style}")
            xa = xa[:, None, :]
            if not self.angle_combine_pairs:
                if reduce_memory: raise NotImplementedError("Angle embedding not implemented with reduce_memory")
                xa = (xa * radial_basis_angle[:, :, None]).reshape(
                    -1, 1, xa.shape[1] * radial_basis_angle.shape[1]
                )

            if self.pair_embedding_key is not None:
                if filtered:
                    ang_pair_src = filter_indices[angle_src]
                    ang_pair_dst = filter_indices[angle_dst]
                else:
                    ang_pair_src = angle_src
                    ang_pair_dst = angle_dst
                ang_pairs = jnp.concatenate((ang_pair_src, ang_pair_dst))

        ##################################################
        ### DIMENSIONS ###
        dim_src = (
            [self.dim_src] * self.nlayers
            if isinstance(self.dim_src, int)
            else self.dim_src
        )
        assert (
            len(dim_src) == self.nlayers
        ), f"dim_src must be an integer or a list of length {self.nlayers}"
        dim_dst = self.dim_dst
        # dim_dst = (
        #     [self.dim_dst] * self.nlayers
        #     if isinstance(self.dim_dst, int)
        #     else self.dim_dst
        # )
        # assert (
        #     len(dim_dst) == self.nlayers
        # ), f"dim_dst must be an integer or a list of length {self.nlayers}"

        if use_angles:
            dim_angle = (
                [self.dim_angle] * self.nlayers
                if isinstance(self.dim_angle, int)
                else self.dim_angle
            )
            assert (
                len(dim_angle) == self.nlayers
            ), f"dim_angle must be an integer or a list of length {self.nlayers}"
            # nmax_angle = [self.nmax_angle]*self.nlayers if isinstance(self.nmax_angle, int) else self.nmax_angle
            # assert len(nmax_angle) == self.nlayers, f"nmax_angle must be an integer or a list of length {self.nlayers}"
        
        initialize_e3 = True
        if self.lmax > 0:
            n_tp = (
                [self.n_tp] * self.nlayers
                if isinstance(self.n_tp, int)
                else self.n_tp
            )
            assert (
                len(n_tp) == self.nlayers
            ), f"n_tp must be an integer or a list of length {self.nlayers}"


        message_passing = (
            [self.message_passing] * self.nlayers
            if isinstance(self.message_passing, bool)
            else self.message_passing
        )
        assert (
            len(message_passing) == self.nlayers
        ), f"message_passing must be a boolean or a list of length {self.nlayers}"

        ##################################################
        ### INITIALIZE PAIR EMBEDDING ###
        if self.pair_embedding_key is not None:
            xij_s,xij_d = jnp.split(nn.Dense(2*dim_dst, name="pair_init_linear")(zi), [dim_dst], axis=-1)
            xij = layer_norm(xij_s[edge_src]*xij_d[edge_dst])

        ##################################################
        if self.keep_all_layers:
            xis = []
        
        ### LOOP OVER LAYERS ###
        for layer in range(self.nlayers):
            ##################################################
            ### COMPACT DESCRIPTORS ###
            si, si_dst = jnp.split(
                nn.Dense(
                    dim_src[layer] + dim_dst,
                    name=f"species_linear_{layer}",
                    use_bias=self.use_bias,
                )(xi),
                [
                    dim_src[layer],
                ],
                axis=-1,
            )

            ##################################################
            if message_passing[layer] or layer == 0:
                ### MESSAGE PASSING ###
                si_mp = si_dst[edge_dst]
            else:
                # if layer == 0:
                #     si_mp = si_dst[edge_dst]
                ### ATTENTION TO SIMULATE MP ###
                Q = nn.Dense(
                    dim_dst * self.att_dim, name=f"queries_{layer}", use_bias=False
                )(si_dst).reshape(-1, dim_dst, self.att_dim)[edge_src]
                K = nn.Dense(
                    dim_dst * self.att_dim, name=f"keys_{layer}", use_bias=False
                )(zi).reshape(-1, dim_dst, self.att_dim)[edge_dst]

                si_mp = (K * Q).sum(axis=-1) / self.att_dim**0.5
                # Vmp = jax.ops.segment_sum(
                #     (KQ * switch)[:, :, None] * Yij, edge_src, species.shape[0]
                # )
                # si_mp = (Vmp[edge_src] * Yij).sum(axis=-1)
                # Q = nn.Dense(
                #     dim_dst * dim_dst, name=f"queries_{layer}", use_bias=False
                # )(si_dst).reshape(-1, dim_dst, dim_dst)
                # si_mp = (
                #     si_mp + jax.vmap(jnp.dot)(Q[edge_src], si_mp) / self.dim_dst**0.5
                # )

            if self.pair_embedding_key is not None:
                si_mp = si_mp + xij

            ##################################################
            ### PAIR EMBEDDING ###
            if reduce_memory:
                Li = jnp.zeros((species.shape[0]* radial_basis.shape[1],si_mp.shape[1]),dtype=si_mp.dtype)
                for i in range(radial_basis.shape[1]):
                    indices = i + edge_src*radial_basis.shape[1]
                    Li = Li.at[indices].add(si_mp*radial_basis[:,i,None])
                Li = Li.reshape(species.shape[0], radial_basis.shape[1]*si_mp.shape[1])
            else:
                Lij = (si_mp[:, None, :] * radial_basis[:, :, None]).reshape(
                    radial_basis.shape[0], si_mp.shape[1] * radial_basis.shape[1]
                )
                ### AGGREGATE PAIR EMBEDDING ###
                Li = jax.ops.segment_sum(Lij, edge_src, species.shape[0])

            ### CONCATENATE EMBEDDING COMPONENTS ###
            components = [si, Li]
            if self.pair_embedding_key is not None:
                if reduce_memory: raise NotImplementedError("Pair embedding not implemented with reduce_memory")
                components_pair = [si[edge_src], xij, Lij]


            ##################################################
            ### ANGLE EMBEDDING ###
            if use_angles and dim_angle[layer]>0:
                si_mp_ang = si_mp[filter_indices] if filtered else si_mp
                if self.angle_combine_pairs:
                    Wa = self.param(
                        f"Wa_{layer}",
                        nn.initializers.normal(
                            stddev=1.0
                            / (si_mp.shape[1] * radial_basis_angle.shape[1]) ** 0.5
                        ),
                        (si_mp.shape[1], radial_basis_angle.shape[1], dim_angle[layer]),
                    )
                    Da = jnp.einsum(
                        "...i,...j,ijk->...k",
                        si_mp_ang,
                        radial_basis_angle,
                        Wa,
                    )

                else:
                    if message_passing[layer] or layer == 0:
                        Da = nn.Dense(dim_angle[layer], name=f"angle_linear_{layer}")(
                            xi
                        )[graph_angle["edge_dst"]]
                    else:
                        Da = nn.Dense(dim_angle[layer], name=f"angle_linear_{layer}")(
                            si_mp_ang
                        )

                Da = Da[angle_dst] * Da[angle_src]
                ## combine pair and angle info
                if reduce_memory:
                    ang_embedding = jnp.zeros((species.shape[0]* Da.shape[-1],xa.shape[-1]),dtype=Da.dtype)
                    for i in range(Da.shape[-1]):
                        indices = i + central_atom*Da.shape[-1]
                        ang_embedding = ang_embedding.at[indices].add(Da[:,i,None]*xa[:,0,:])
                    ang_embedding = ang_embedding.reshape(species.shape[0], xa.shape[-1]*Da.shape[-1])
                else:
                    radang = (xa * Da[:, :, None]).reshape(
                        (-1, Da.shape[1] * xa.shape[2])
                    )
                    ### AGGREGATE  ANGLE EMBEDDING ###
                    ang_embedding = jax.ops.segment_sum(
                        radang, central_atom, species.shape[0]
                    )
                    

                components.append(ang_embedding)

                if self.pair_embedding_key is not None:
                    ang_ij = jax.ops.segment_sum(
                        jnp.concatenate((radang, radang)),
                        ang_pairs,
                        edge_src.shape[0],
                    )
                    components_pair.append(ang_ij)
            
            ##################################################
            ### EQUIVARIANT EMBEDDING ###
            if self.lmax > 0 and n_tp[layer] >= 0:
                if initialize_e3 or not message_passing[layer]:
                    Vij = Yij
                elif self.edge_tp:
                    Vij = FilteredTensorProduct(
                            self.lmax, self.lmax, name=f"edge_tp_{layer}",ignore_parity=self.ignore_irreps_parity,weights_by_channel=False
                        )(Vi[edge_dst], Yij)
                else:
                    Vij = Vi[edge_dst]

                ### compute channel weights
                dim_wij = self.nchannels_l
                if self.resolve_wij_l:
                    dim_wij=self.nchannels_l*(self.lmax+1)

                eij = Lij if self.pair_embedding_key is None else jnp.concatenate([Lij,xij*switch],axis=-1)
                wij = nn.Dense(
                        dim_wij, name=f"e3_channel_{layer}", use_bias=False
                    )(eij)
                if self.resolve_wij_l:
                    wij = jnp.repeat(wij.reshape(-1,self.nchannels_l,self.lmax+1),nrep_l,axis=-1)
                else:
                    wij = wij[:,:,None]
                
                ### aggregate equivariant messages
                drhoi = jax.ops.segment_sum(
                    wij * Vij,
                    edge_src,
                    species.shape[0],
                )

                Vi0 = []
                if initialize_e3:
                    rhoi = drhoi
                    Vi = ChannelMixingE3(
                        self.lmax,
                        self.nchannels_l,
                        self.nchannels_l,
                        name=f"e3_initial_mixing_{layer}",
                    )(rhoi)
                    # assert n_tp[layer] > 0, "n_tp must be > 0 for the first equivariant layer."
                else:
                    rhoi = rhoi + drhoi
                    # if message_passing[layer]:
                        # Vi0.append(drhoi[:, :, 0])
                initialize_e3 = False
                if n_tp[layer] > 0:
                    for itp in range(n_tp[layer]):
                        dVi = FilteredTensorProduct(
                            self.lmax, self.lmax, name=f"tensor_product_{layer}_{itp}",ignore_parity=self.ignore_irreps_parity,weights_by_channel=False
                        )(rhoi, Vi)
                        Vi = ChannelMixing(
                                self.lmax,
                                self.nchannels_l,
                                self.nchannels_l,
                                name=f"tp_mixing_{layer}_{itp}",
                            )(Vi + dVi)
                        Vi0.append(dVi[:, :, 0])
                    Vi0 = jnp.concatenate(Vi0, axis=-1)
                    components.append(Vi0)

                if self.pair_embedding_key is not None:
                    Vij = Vi[edge_src]*Yij
                    Vij0 = [Vij[...,0]]
                    for l in range(1,self.lmax+1):
                        Vij0.append(Vij[...,l**2:(l+1)**2].sum(axis=-1))
                    Vij0 = jnp.concatenate(Vij0,axis=-1)
                    components_pair.append(Vij0)

            ##################################################
            ### CONCATENATE EMBEDDING COMPONENTS ###
            if do_lode and nchannels_lode[layer] > 0:
                zj = nn.Dense(dim_lr[layer], use_bias=False, name=f"LODE_{layer}")(xi)
                if nextra_powers > 0:
                    zj_extra = zj[:,:nextra_powers*nchannels_lode[layer]].reshape(
                        (species.shape[0],nchannels_lode[layer], nextra_powers)
                    )
                    zj = zj[:,nextra_powers*nchannels_lode[layer]:]
                    xi_lr_extra = jax.ops.segment_sum(
                        eij_lr_extra * zj_extra[edge_dst_lr], edge_src_lr, species.shape[0]
                    )
                    components.append(xi_lr_extra.reshape(species.shape[0],-1))

                if equivariant_lode:
                    zj = zj.reshape(
                        (species.shape[0], nchannels_lode[layer], lmax_lr + 1)
                    ).repeat(nrep_lr, axis=-1)
                xi_lr = jax.ops.segment_sum(
                    eij_lr * zj[edge_dst_lr], edge_src_lr, species.shape[0]
                )
                if equivariant_lode:
                    assert self.lode_use_field_norm or self.lode_multipole_interaction, "equivariant LODE requires field norm or multipole interaction"
                    if self.lode_multipole_interaction:
                        if initialize_e3:
                            raise ValueError("equivariant LODE used before local equivariants initialized")
                        size_l_lr = (lmax_lr+1)**2
                        if self.lode_direct_multipoles:
                            assert nchannels_lode[layer] <= self.nchannels_l
                            Mi = Vi[:, : nchannels_lode[layer], :size_l_lr]
                        else:
                            Mi = ChannelMixingE3(
                                lmax_lr,
                                self.nchannels_l,
                                nchannels_lode[layer],
                                name=f"e3_LODE_{layer}",
                            )(Vi[...,:size_l_lr])
                        Mi_lr = Mi * xi_lr
                    components.append(xi_lr[:, :, 0])
                    if self.lode_use_field_norm and self.lode_equi_full_combine:
                        xi_lr1 = ChannelMixing(
                            lmax_lr,
                            nchannels_lode[layer],
                            nchannels_lode[layer],
                            name=f"LODE_mixing_{layer}",
                        )(xi_lr)
                    norm = 1.
                    for l in range(1, lmax_lr + 1):
                        if self.lode_normalize_l:
                            norm = 1. / (2 * l + 1)
                        if self.lode_multipole_interaction:
                            components.append(Mi_lr[:, :, l**2 : (l + 1) ** 2].sum(axis=-1)*norm)

                        if self.lode_use_field_norm:
                            if self.lode_equi_full_combine:
                                components.append((xi_lr[:,:,l**2 : (l + 1) ** 2]*xi_lr1[:,:,l**2 : (l + 1) ** 2]).sum(axis=-1)*norm)
                            else:
                                components.append(
                                    ((xi_lr[:, :, l**2 : (l + 1) ** 2]) ** 2).sum(axis=-1)*norm
                                )
                else:
                    components.append(xi_lr)

            dxi = jnp.concatenate(components, axis=-1)

            ##################################################
            ### CONCATENATE PAIR EMBEDDING COMPONENTS ###
            if self.pair_embedding_key is not None:
                dxij = jnp.concatenate(components_pair, axis=-1)

            ##################################################
            ### MIX AND APPLY NONLINEARITY ###
            if self.block_index_key is not None:
                block_index = inputs[self.block_index_key]
                dxi = actmix(BlockIndexNet(
                        output_dim=self.dim,
                        hidden_neurons=self.mixing_hidden,
                        activation=self.activation,
                        name=f"dxi_{layer}",
                        use_bias=self.use_bias,
                        kernel_init=kernel_init,
                    )((species,dxi, block_index))
                )
            else:
                dxi = actmix(
                    FullyConnectedNet(
                        [*self.mixing_hidden, self.dim],
                        activation=self.activation,
                        name=f"dxi_{layer}",
                        use_bias=self.use_bias,
                        kernel_init=kernel_init,
                    )(dxi)
                )

            if self.pair_embedding_key is not None:
                ### UPDATE PAIR EMBEDDING ###
                # dxij = tssr3(nn.Dense(dim_dst, name=f"dxij_{layer}",use_bias=False)(dxij))
                dxij = actmix(
                    FullyConnectedNet(
                        [*self.pair_mixing_hidden, dim_dst],
                        activation=self.activation,
                        name=f"dxij_{layer}",
                        use_bias=False,
                        kernel_init=kernel_init,
                    )(dxij)
                )
                xij = layer_norm(xij + dxij)

            ##################################################
            ### UPDATE EMBEDDING ###
            if layer == 0 and not (self.species_init or self.charge_embedding):
                xi = layer_norm(dxi)
            else:
                ### FORGET GATE ###
                R = jax.nn.sigmoid(
                    self.param(
                        f"retention_{layer}",
                        nn.initializers.normal(),
                        (xi.shape[-1],),
                    )
                )
                xi = layer_norm(R[None, :] * xi + dxi)

            if self.keep_all_layers:
                xis.append(xi)

        embedding_key = (
            self.embedding_key if self.embedding_key is not None else self.name
        )
        output = {
            **inputs,
            embedding_key: xi,
        }
        if self.lmax > 0:
            output[embedding_key + "_tensor"] = Vi
        if self.keep_all_layers:
            output[embedding_key + "_layers"] = jnp.stack(xis, axis=1)
        if self.charge_embedding:
            output[embedding_key + "_charge"] = charge_embedding
        if self.pair_embedding_key is not None:
            output[self.pair_embedding_key] = xij
        return output
